__all__ = ('API_ENDPOINT', 'CDN_ENDPOINT', 'DISCORD_ENDPOINT', 'INVITE_URL_RP', 'MESSAGE_JUMP_URL_RP',
    'STATUS_ENDPOINT', 'VALID_ICON_FORMATS', 'VALID_ICON_FORMATS_EXTENDED', 'VALID_ICON_MEDIA_TYPES',
    'VALID_ICON_MEDIA_TYPES_EXTENDED', 'VALID_STICKER_IMAGE_MEDIA_TYPES', 'is_media_url')

import re

from ...env import CUSTOM_API_ENDPOINT, CUSTOM_CDN_ENDPOINT, CUSTOM_DISCORD_ENDPOINT, CUSTOM_STATUS_ENDPOINT, \
    API_VERSION

from ...backend.export import export, include

from ..bases import ICON_TYPE_NONE, ICON_TYPE_STATIC

ChannelGuildBase = include('ChannelGuildBase')
StickerFormat = include('StickerFormat')

API_ENDPOINT = f'https://discord.com/api/v{API_VERSION}' if (CUSTOM_API_ENDPOINT is None) else CUSTOM_API_ENDPOINT
CDN_ENDPOINT = 'https://cdn.discordapp.com' if (CUSTOM_CDN_ENDPOINT is None) else CUSTOM_CDN_ENDPOINT
DISCORD_ENDPOINT = 'https://discord.com' if (CUSTOM_DISCORD_ENDPOINT is None) else CUSTOM_DISCORD_ENDPOINT
STATUS_ENDPOINT = 'https://status.discord.com/api/v2' if (CUSTOM_STATUS_ENDPOINT is None) else CUSTOM_STATUS_ENDPOINT

del CUSTOM_API_ENDPOINT, CUSTOM_CDN_ENDPOINT, CUSTOM_DISCORD_ENDPOINT, CUSTOM_STATUS_ENDPOINT, API_VERSION

VALID_ICON_SIZES = frozenset((
    *( 1<<x    for x in range(4, 13)),
    *((1<<x)*3 for x in range(9, 11)),
    *((1<<x)*5 for x in range(2,  9)),
))

VALID_ICON_FORMATS = frozenset(('jpg', 'jpeg', 'png', 'webp'))
VALID_ICON_FORMATS_EXTENDED = frozenset((*VALID_ICON_FORMATS, 'gif',))

VALID_ICON_MEDIA_TYPES = frozenset(('image/jpeg', 'image/png', 'image/webp'))
VALID_ICON_MEDIA_TYPES_EXTENDED = frozenset(('image/gif', *VALID_ICON_MEDIA_TYPES))

VALID_STICKER_IMAGE_MEDIA_TYPES = frozenset(('image/png', 'application/json'))

STYLE_PATTERN = re.compile('(^shield$)|(^banner[1-4]$)')

MESSAGE_JUMP_URL_RP = re.compile('(?:https://)?discord(?:app)?.com/channels/(?:(\d{7,21})|@me)/(\d{7,21})/(\d{7,21})')
export(MESSAGE_JUMP_URL_RP, 'MESSAGE_JUMP_URL_RP')

#returns a URL that allows the client to jump to this message
#guild is guild's id, or @me if there is no guild
def message_jump_url(message):
    """
    Returns a jump url to the message. If the message's channel is a partial guild channel, returns `None`.
    
    Parameters
    ----------
    message : ``Message``
        The respective message.
    
    Returns
    -------
    url : `None` or `str`
    """
    channel_id = message.channel_id
    guild_id = message.guild_id
    if guild_id:
        guild_id = str(guild_id)
    else:
        guild_id = '@me'
    
    return f'{DISCORD_ENDPOINT}/channels/{guild_id}/{channel_id}/{message.id}'

CDN_RP = re.compile(
    'https://(?:'
        'cdn\.discordapp\.com|'
        'discord\.com|'
        '(?:'
            'images-ext-\d+|'
            'media'
        ')\.discordapp\.net'
    ')/'
)

def is_cdn_url(url):
    """
    Returns whether the given url a Discord content delivery network url.
    
    Parameters
    ----------
    url : `str`
        The url to check.
    
    Returns
    -------
    is_cdn_url : `bool`
    
    Examples
    --------
    Icons: `https://cdn.discordapp.com/...`
    Assets: `https://discord.com/...`
    Proxy service: `https://images-ext-1.discordapp.net/...`
    Attachments: `https://media.discordapp.net/...`
    ```
    """
    return (CDN_RP.match(url) is not None)


def is_media_url(url):
    """
    Returns whether the given url uses the discord's media content delivery network.
    
    Parameters
    ----------
    url : `str`
        The url to check.
    
    Returns
    -------
    is_media_url : `bool`
    """
    return url.startswith('https://media.discordapp.net/')


def guild_icon_url(guild):
    """
    Returns the guild's icon's image's url. If the guild has no icon, then returns `None`.
    
    Parameters
    ----------
    guild : ``Guild`` or ``GuildPreview``
        The respective guild.
    
    Returns
    -------
    url : `None` or `str`
    """
    icon_type = guild.icon_type
    if icon_type is ICON_TYPE_NONE:
        return None
    
    if icon_type is ICON_TYPE_STATIC:
        prefix = ''
        ext = 'png'
    else:
        prefix = 'a_'
        ext = 'gif'
    
    return f'{CDN_ENDPOINT}/icons/{guild.id}/{prefix}{guild.icon_hash:0>32x}.{ext}'


def guild_icon_url_as(guild, ext=None, size=None):
    """
    Returns the guild's icon's url. If the guild has no icon, then returns `None`.
    
    Parameters
    ----------
    guild : ``Guild`` or ``GuildPreview``
        The respective guild.
    ext : `str`, Optional
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`. If the guild has
        animated icon, it can `'gif'` as well.
    size : `int`, Optional
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None` or `str`
    
    Raises
    ------
    ValueError
        If `ext` or `size` was not passed as any of the expected values.
    """
    icon_type = guild.icon_type
    if icon_type is ICON_TYPE_NONE:
        return None
    
    if size is None:
        end = ''
    elif size in VALID_ICON_SIZES:
        end = f'?size={size}'
    else:
        raise ValueError(f'Size must be in {sorted(VALID_ICON_SIZES)!r}, got {size}.')
    
    if ext is None:
        if icon_type is ICON_TYPE_STATIC:
            prefix = ''
            ext = 'png'
        else:
            prefix = 'a_'
            ext = 'gif'
    
    else:
        if icon_type is ICON_TYPE_STATIC:
            if ext not in VALID_ICON_FORMATS:
                raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS}, got {ext!r}.')
            prefix = ''
        else:
            if ext not in VALID_ICON_FORMATS_EXTENDED:
                raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS_EXTENDED}, got {ext!r}.')
            prefix = 'a_'
    
    return f'{CDN_ENDPOINT}/icons/{guild.id}/{prefix}{guild.icon_hash:0>32x}.{ext}{end}'


def guild_invite_splash_url(guild):
    """
    Returns the guild's invite splash's image's url. If the guild has no invite splash, then returns `None`.
    
    Parameters
    ----------
    guild : ``Guild`` or ``GuildPreview``
        The respective guild.
    
    Returns
    -------
    url : `None` or `str`
    """
    icon_type = guild.invite_splash_type
    if icon_type is ICON_TYPE_NONE:
        return None
    
    if icon_type is ICON_TYPE_STATIC:
        prefix = ''
        ext = 'png'
    else:
        prefix = 'a_'
        ext = 'gif'
    
    return f'{CDN_ENDPOINT}/splashes/{guild.id}/{prefix}{guild.invite_splash_hash:0>32x}.{ext}'


def guild_invite_splash_url_as(guild, ext=None, size=None):
    """
    Returns the guild's invite splash's image's url. If the guild has no invite splash, then returns `None`.
    
    Parameters
    ----------
    guild : ``Guild`` or ``GuildPreview``
        The respective guild.
    ext : `str`, Optional
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
    size : `int`, Optional
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None` or `str`
    
    Raises
    ------
    ValueError
        If `ext` or `size` was not passed as any of the expected values.
    """
    icon_type = guild.invite_splash_type
    if icon_type is ICON_TYPE_NONE:
        return None
    
    if size is None:
        end = ''
    elif size in VALID_ICON_SIZES:
        end = f'?size={size}'
    else:
        raise ValueError(f'Size must be in {sorted(VALID_ICON_SIZES)!r}, got {size}.')
    
    if ext is None:
        if icon_type is ICON_TYPE_STATIC:
            prefix = ''
            ext = 'png'
        else:
            prefix = 'a_'
            ext = 'gif'
    
    else:
        if icon_type is ICON_TYPE_STATIC:
            if ext not in VALID_ICON_FORMATS:
                raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS}, got {ext!r}.')
            prefix = ''
        else:
            if ext not in VALID_ICON_FORMATS_EXTENDED:
                raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS_EXTENDED}, got {ext!r}.')
            prefix = 'a_'
    
    return f'{CDN_ENDPOINT}/splashes/{guild.id}/{prefix}{guild.invite_splash_hash:0>32x}.{ext}{end}'


def guild_discovery_splash_url(guild):
    """
    Returns the guild's discovery splash's image's url. If the guild has no discovery splash, then returns `None`.
    
    Parameters
    ----------
    guild : ``Guild`` or ``GuildPreview``
        The respective guild.
    
    Returns
    -------
    url : `None` or `str`
    """
    icon_type = guild.discovery_splash_type
    if icon_type is ICON_TYPE_NONE:
        return None
    
    if icon_type is ICON_TYPE_STATIC:
        prefix = ''
        ext = 'png'
    else:
        prefix = 'a_'
        ext = 'gif'
    
    return f'{CDN_ENDPOINT}/discovery-splashes/{guild.id}/{prefix}{guild.discovery_splash_hash:0>32x}.{ext}'


def guild_discovery_splash_url_as(guild, ext=None, size=None):
    """
    Returns the guild's discovery splash's image's url. If the guild has no discovery splash, then returns `None`.
    
    Parameters
    ----------
    guild : ``Guild`` or ``GuildPreview``
        The respective guild.
    ext : `str`, Optional
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
    size : `int`, Optional
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None` or `str`
    
    Raises
    ------
    ValueError
        If `ext` or `size` was not passed as any of the expected values.
    """
    icon_type = guild.discovery_splash_type
    if icon_type is ICON_TYPE_NONE:
        return None
    
    if size is None:
        end = ''
    elif size in VALID_ICON_SIZES:
        end = f'?size={size}'
    else:
        raise ValueError(f'Size must be in {sorted(VALID_ICON_SIZES)!r}, got {size}.')
    
    if ext is None:
        if icon_type is ICON_TYPE_STATIC:
            prefix = ''
            ext = 'png'
        else:
            prefix = 'a_'
            ext = 'gif'
    
    else:
        if icon_type is ICON_TYPE_STATIC:
            if ext not in VALID_ICON_FORMATS:
                raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS}, got {ext!r}.')
            prefix = ''
        else:
            if ext not in VALID_ICON_FORMATS_EXTENDED:
                raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS_EXTENDED}, got {ext!r}.')
            prefix = 'a_'
    
    return f'{CDN_ENDPOINT}/discovery-splashes/{guild.id}/{prefix}{guild.discovery_splash_hash:0>32x}.{ext}{end}'


def guild_banner_url(guild):
    """
    Returns the guild's banner's image's url. If the guild has no banner, then returns `None`.
    
    Parameters
    ----------
    guild : ``Guild``
        The respective guild.
    
    Returns
    -------
    url : `None` or `str`
    """
    icon_type = guild.banner_type
    if icon_type is ICON_TYPE_NONE:
        return None
    
    if icon_type is ICON_TYPE_STATIC:
        prefix = ''
        ext = 'png'
    else:
        prefix = 'a_'
        ext = 'gif'
    
    return f'{CDN_ENDPOINT}/banners/{guild.id}/{prefix}{guild.banner_hash:0>32x}.{ext}'


def guild_banner_url_as(guild, ext=None, size=None):
    """
    Returns the guild's banner's image's url. If the guild has no banner, then returns `None`.
    
    Parameters
    ----------
    guild : ``Guild``
        The respective guild.
    ext : `str`, Optional
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`, `'gif'`.
    size : `int`, Optional
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None` or `str`
    
    Raises
    ------
    ValueError
        If `ext` or `size` was not passed as any of the expected values.
    """
    icon_type = guild.banner_type
    if icon_type is ICON_TYPE_NONE:
        return None
    
    if size is None:
        end = ''
    elif size in VALID_ICON_SIZES:
        end = f'?size={size}'
    else:
        raise ValueError(f'Size must be in {sorted(VALID_ICON_SIZES)!r}, got {size}.')
    
    if ext is None:
        if icon_type is ICON_TYPE_STATIC:
            prefix = ''
            ext = 'png'
        else:
            prefix = 'a_'
            ext = 'gif'
    
    else:
        if icon_type is ICON_TYPE_STATIC:
            if ext not in VALID_ICON_FORMATS:
                raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS}, got {ext!r}.')
            prefix = ''
        else:
            if ext not in VALID_ICON_FORMATS_EXTENDED:
                raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS_EXTENDED}, got {ext!r}.')
            prefix = 'a_'

    return f'{CDN_ENDPOINT}/banners/{guild.id}/{prefix}{guild.banner_hash:0>32x}.{ext}{end}'


def guild_widget_url(guild, style='shield'):
    """
    Returns the guild's widget image's url in `.png` format.
    
    Parameters
    ----------
    guild : ``Guild`` or ``GuildPreview``
        The respective guild.
    style : `str`
        The widget image's style. Can be any of: `'shield'`, `'banner1'`, `'banner2'`, `'banner3'`, `'banner4'`.
    
    Returns
    -------
    url : `str`
    
    Raises
    ------
    ValueError
        If `style` was not passed as any of the expected values.
    """
    if STYLE_PATTERN.match(style) is None:
        raise ValueError(f'Invalid style: {style!r}')
    
    return f'{API_ENDPOINT}/guilds/{guild.id}/widget.png?style={style}'


def guild_widget_json_url(guild):
    """
    Returns an url to request a ``Guild``'s widget data.
    
    Parameters
    ----------
    guild : ``Guild`` or ``GuildPreview``
        The respective guild.
    
    Returns
    -------
    url : `str`
    """
    return  f'{API_ENDPOINT}/guilds/{guild.id}/widget.json'


def channel_group_icon_url(channel):
    """
    Returns the group channel's icon's image's url. If the channel has no icon, then returns `None`.
    
    Parameters
    ----------
    channel : ``ChannelGroup``
        The respective channel.
    
    Returns
    -------
    url : `None` or `str`
    """
    icon_type = channel.icon_type
    if icon_type is ICON_TYPE_NONE:
        return None
    
    if icon_type is ICON_TYPE_STATIC:
        prefix = ''
        ext = 'png'
    else:
        prefix = 'a_'
        ext = 'gif'
    
    return f'{CDN_ENDPOINT}/channel-icons/{channel.id}/{prefix}{channel.icon_hash:0>32x}.{ext}'
    
    
def channel_group_icon_url_as(channel, ext=None, size=None):
    """
    Returns the group channel's icon's image's url. If the channel has no icon, then returns `None`.
    
    Parameters
    ----------
    channel : ``ChannelGroup``
        The respective channel.
    ext : `str`, Optional
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
    size : `int`, Optional
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None` or `str`
    
    Raises
    ------
    ValueError
        If `ext` or `size` was not passed as any of the expected values.
    """
    icon_type = channel.icon_type
    if icon_type is ICON_TYPE_NONE:
        return None
    
    if size is None:
        end = ''
    elif size in VALID_ICON_SIZES:
        end = f'?size={size}'
    else:
        raise ValueError(f'Size must be in {sorted(VALID_ICON_SIZES)!r}, got {size}.')
    
    if ext is None:
        if icon_type is ICON_TYPE_STATIC:
            prefix = ''
            ext = 'png'
        else:
            prefix = 'a_'
            ext = 'gif'
    
    else:
        if icon_type is ICON_TYPE_STATIC:
            if ext not in VALID_ICON_FORMATS:
                raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS}, got {ext!r}.')
            prefix = ''
        else:
            if ext not in VALID_ICON_FORMATS_EXTENDED:
                raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS_EXTENDED}, got {ext!r}.')
            prefix = 'a_'
    
    return f'{CDN_ENDPOINT}/channel-icons/{channel.id}/{prefix}{channel.icon_hash:0>32x}.{ext}{end}'


def emoji_url(emoji):
    """
    Returns the emoji's image's url. If the emoji is unicode emoji, then returns `None` instead.
    
    Parameters
    ----------
    emoji : ``Emoji``
        The respective emoji.
    
    Returns
    -------
    url : `None` or `str`
    """
    if emoji.is_unicode_emoji():
        return None
    
    if emoji.animated:
         ext = 'gif'
    else:
         ext = 'png'
        
    return f'{CDN_ENDPOINT}/emojis/{emoji.id}.{ext}'


def emoji_url_as(emoji, ext=None, size=None):
    """
    Returns the emoji's image's url. If the emoji is unicode emoji, then returns `None` instead.
    
    Parameters
    ----------
    emoji : ``Emoji``
        The respective emoji.
    ext : `str`, Optional
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`. If emoji is
        animated, it can `'gif'` as well.
    size : `int`, Optional
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None` or `str`
    
    Raises
    ------
    ValueError
        If `ext` or `size` was not passed as any of the expected values.
    """
    if emoji.is_unicode_emoji():
        return None

    if size is None:
        end = ''
    elif size in VALID_ICON_SIZES:
        end = f'?size={size}'
    else:
        raise ValueError(f'Size must be in {sorted(VALID_ICON_SIZES)!r}, got {size}.')
    
    if ext is None:
        if emoji.animated:
            ext = 'gif'
        else:
            ext = 'png'
    else:
        if emoji.animated:
            if ext not in VALID_ICON_FORMATS_EXTENDED:
                raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS_EXTENDED}, got {ext!r}.')
        else:
            if ext not in VALID_ICON_FORMATS:
                raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS}, got {ext!r}.')
    
    return f'{CDN_ENDPOINT}/emojis/{emoji.id}.{ext}{end}'


def webhook_url(webhook):
    """
    Returns the webhook's url.
    
    Parameters
    ----------
    webhook : `Webhook``
        The respective webhook.
    
    Returns
    -------
    url : `str`
    """
    return f'{API_ENDPOINT}/webhooks/{webhook.id}/{webhook.token}'


WEBHOOK_URL_PATTERN = re.compile(
    '(?:https://)?discord(?:app)?.com/api/(?:v\d+/)?webhooks/([0-9]{17,21})/([a-zA-Z0-9\.\-\_%]{60,68})(?:/.*)?'
)


def invite_url(invite):
    """
    Returns the invite's url.
    
    Parameters
    ----------
    invite : ``Invite``
        The respective invite.
    
    Returns
    -------
    url : `str`
    """
    return f'http://discord.gg/{invite.code}'


INVITE_URL_RP = re.compile('(?:https?://)?discord(?:\.gg|(?:app)?\.com/invite)/([a-zA-Z0-9-]+)')


def activity_asset_image_large_url(activity):
    """
    Returns the activity's large asset image's url. If the activity has no large asset image, then returns `None`.
    
    Parameters
    ----------
    activity : ``ActivityRich``
        The respective activity.
    
    Returns
    -------
    url : `None` or `str`
    """
    application_id = activity.application_id
    if not application_id:
        return None
    
    assets = activity.assets
    if assets is None:
        return None
    
    image_large = assets.image_large
    if image_large is None:
        return None
    
    return f'{CDN_ENDPOINT}/app-assets/{application_id}/{image_large}.png'


def activity_asset_image_large_url_as(activity, ext=None, size=None):
    """
    Returns the activity's large asset image's url. If the activity has no large asset image, then returns `None`.

    
    Parameters
    ----------
    activity : ``ActivityRich``
        The respective activity.
    ext : `str`, Optional
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
    size : `int`, Optional
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None` or `str`
    
    Raises
    ------
    ValueError
        If `ext` or `size` was not passed as any of the expected values.
    """
    application_id = activity.application_id
    if not application_id:
        return None

    assets = activity.assets
    if assets is None:
        return None
    
    image_large = assets.image_large
    if image_large is None:
        return None

    if size is None:
        end = ''
    elif size in VALID_ICON_SIZES:
        end = f'?size={size}'
    else:
        raise ValueError(f'Size must be in {sorted(VALID_ICON_SIZES)!r}, got {size}.')

    if ext not in VALID_ICON_FORMATS:
        raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS}, got {ext!r}.')

    return f'{CDN_ENDPOINT}/app-assets/{application_id}/{image_large}.{ext}{end}'


def activity_asset_image_small_url(activity):
    """
    Returns the activity's small asset image's url. If the activity has no small asset image, then returns `None`.
    
    Parameters
    ----------
    activity : ``ActivityRich``
        The respective activity.
    
    Returns
    -------
    url : `None` or `str`
    """
    application_id = activity.application_id
    if not application_id:
        return None
    
    assets = activity.assets
    if assets is None:
        return None
    
    image_small = assets.image_small
    if image_small is None:
        return None
    
    return f'{CDN_ENDPOINT}/app-assets/{application_id}/{image_small}.png'


def activity_asset_image_small_url_as(activity, ext=None, size=None):
    """
    Returns the activity's small asset image's url. If the activity has no small asset image, then returns `None`.
    
    Parameters
    ----------
    activity : ``ActivityRich``
        The respective activity.
    ext : `str`, Optional
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
    size : `int`, Optional
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None` or `str`
    
    Raises
    ------
    ValueError
        If `ext` or `size` was not passed as any of the expected values.
    """
    application_id = activity.application_id
    if not application_id:
        return None
    
    assets = activity.assets
    if assets is None:
        return None
    
    image_small = assets.image_small
    if image_small is None:
        return None
    
    if size is None:
        end = ''
    elif size in VALID_ICON_SIZES:
        end = f'?size={size}'
    else:
        raise ValueError(f'Size must be in {sorted(VALID_ICON_SIZES)!r}, got {size}.')
    
    if ext not in VALID_ICON_FORMATS:
        raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS}, got {ext!r}.')
    
    return f'{CDN_ENDPOINT}/app-assets/{application_id}/{image_small}.{ext}{end}'


def user_avatar_url(user):
    """
    Returns the user's avatar's url. If the user has no avatar, then returns it's default avatar's url.
    
    Parameters
    ----------
    user : ``UserBase``
        The respective user.
    
    Returns
    -------
    url : `None` or `str`
    """
    icon_type = user.avatar_type
    if icon_type is ICON_TYPE_NONE:
        return user.default_avatar.url
    
    if icon_type is ICON_TYPE_STATIC:
        prefix = ''
        ext = 'png'
    else:
        prefix = 'a_'
        ext = 'gif'
    
    return f'{CDN_ENDPOINT}/avatars/{user.id}/{prefix}{user.avatar_hash:0>32x}.{ext}'


def user_avatar_url(user):
    """
    Returns the user's avatar's url. If the user has no avatar, then returns it's default avatar's url.
    
    Parameters
    ----------
    user : ``UserBase``
        The respective user.
    
    Returns
    -------
    url : `None` or `str`
    """
    icon_type = user.avatar_type
    if icon_type is ICON_TYPE_NONE:
        return user.default_avatar.url
    
    if icon_type is ICON_TYPE_STATIC:
        prefix = ''
        ext = 'png'
    else:
        prefix = 'a_'
        ext = 'gif'
    
    return f'{CDN_ENDPOINT}/avatars/{user.id}/{prefix}{user.avatar_hash:0>32x}.{ext}'


def user_avatar_url_as(user, ext=None, size=None):
    """
    Returns the user's avatar's url. If the user has no avatar, then returns it's default avatar's url.
    
    Parameters
    ----------
    user : ``UserBase``
        The respective user.
    ext : `str`, Optional
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`. If the user has
        animated avatar, it can `'gif'` as well.
    size : `int`, Optional
        The preferred minimal size of the avatar's url.
    
    Returns
    -------
    url : `None` or `str`
    
    Raises
    ------
    ValueError
        If `ext` or `size` was not passed as any of the expected values.
    """
    icon_type = user.avatar_type
    if icon_type is ICON_TYPE_NONE:
        return user.default_avatar.url
    
    if size is None:
        end = ''
    elif size in VALID_ICON_SIZES:
        end = f'?size={size}'
    else:
        raise ValueError(f'Size must be in {sorted(VALID_ICON_SIZES)!r}, got {size}.')
    
    if ext is None:
        if icon_type is ICON_TYPE_STATIC:
            prefix = ''
            ext = 'png'
        else:
            prefix = 'a_'
            ext = 'gif'
    
    else:
        if icon_type is ICON_TYPE_STATIC:
            if ext not in VALID_ICON_FORMATS:
                raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS}, got {ext!r}.')
            prefix = ''
        else:
            if ext not in VALID_ICON_FORMATS_EXTENDED:
                raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS_EXTENDED}, got {ext!r}.')
            prefix = 'a_'
    
    return f'{CDN_ENDPOINT}/avatars/{user.id}/{prefix}{user.avatar_hash:0>32x}.{ext}{end}'


def user_banner_url(user):
    """
    Returns the user's banner's url. If the user has no banner, then returns `None`.
    
    Parameters
    ----------
    user : ``UserBase``
        The respective user.
    
    Returns
    -------
    url : `None` or `str`
    """
    icon_type = user.banner_type
    if icon_type is ICON_TYPE_NONE:
        return None
    
    if icon_type is ICON_TYPE_STATIC:
        prefix = ''
        ext = 'png'
    else:
        prefix = 'a_'
        ext = 'gif'
    
    return f'{CDN_ENDPOINT}/banners/{user.id}/{prefix}{user.banner_hash:0>32x}.{ext}'


def user_banner_url_as(user, ext=None, size=None):
    """
    Returns the user's banner's url. If the user has no banner, then returns `None`.
    
    Parameters
    ----------
    user : ``UserBase``
        The respective user.
    ext : `str`, Optional
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`. If the user has
        animated banner, it can `'gif'` as well.
    size : `int`, Optional
        The preferred minimal size of the avatar's url.
    
    Returns
    -------
    url : `None` or `str`
    
    Raises
    ------
    ValueError
        If `ext` or `size` was not passed as any of the expected values.
    """
    icon_type = user.banner_type
    if icon_type is ICON_TYPE_NONE:
        return None
    
    if size is None:
        end = ''
    elif size in VALID_ICON_SIZES:
        end = f'?size={size}'
    else:
        raise ValueError(f'Size must be in {sorted(VALID_ICON_SIZES)!r}, got {size}.')
    
    if ext is None:
        if icon_type is ICON_TYPE_STATIC:
            prefix = ''
            ext = 'png'
        else:
            prefix = 'a_'
            ext = 'gif'
    
    else:
        if icon_type is ICON_TYPE_STATIC:
            if ext not in VALID_ICON_FORMATS:
                raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS}, got {ext!r}.')
            prefix = ''
        else:
            if ext not in VALID_ICON_FORMATS_EXTENDED:
                raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS_EXTENDED}, got {ext!r}.')
            prefix = 'a_'
    
    return f'{CDN_ENDPOINT}/banners/{user.id}/{prefix}{user.banner_hash:0>32x}.{ext}{end}'


def user_avatar_url_for(user, guild):
    """
    Returns the user's guild specific avatar. If the user has no guild specific avatar, returns `None`.
    
    Parameters
    ----------
    user : ``UserBase``
        The Respective user.
    guild : ``Guild``
        The respective guild.
    
    Returns
    -------
    url : `None` or `str`
    """
    if guild is None:
        return None
    
    try:
        guild_profile = user.guild_profiles[guild.id]
    except KeyError:
        return None
    
    icon_type = guild_profile.avatar_type
    if icon_type is ICON_TYPE_NONE:
        return None
    
    if icon_type is ICON_TYPE_STATIC:
        prefix = ''
        ext = 'png'
    else:
        prefix = 'a_'
        ext = 'gif'
    
    return f'{CDN_ENDPOINT}/guilds/{guild.id}/users/{user.id}/avatars/{prefix}{guild_profile.avatar_hash:0>32x}.{ext}'


def user_avatar_url_for_as(user, guild, ext=None, size=None):
    """
    Returns the user's guild specific avatar. If the user has no avatar, then returns it's default avatar's url.
    
    Parameters
    ----------
    user : ``UserBase``
        The Respective user.
    guild : ``Guild``
        The respective guild.
    ext : `str`, Optional
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`. If the user has
        animated avatar, it can `'gif'` as well.
    size : `int`, Optional
        The preferred minimal size of the avatar's url.
    
    Returns
    -------
    url : `None` or `str`
    
    Raises
    ------
    ValueError
        If `ext` or `size` was not passed as any of the expected values.
    """
    if guild is None:
        return None
    
    try:
        guild_profile = user.guild_profiles[guild.id]
    except KeyError:
        return None
    
    icon_type = guild_profile.avatar_type
    if icon_type is ICON_TYPE_NONE:
        return None
    
    if size is None:
        end = ''
    elif size in VALID_ICON_SIZES:
        end = f'?size={size}'
    else:
        raise ValueError(f'Size must be in {sorted(VALID_ICON_SIZES)!r}, got {size}.')
    
    if ext is None:
        if icon_type is ICON_TYPE_STATIC:
            prefix = ''
            ext = 'png'
        else:
            prefix = 'a_'
            ext = 'gif'
    
    else:
        if icon_type is ICON_TYPE_STATIC:
            if ext not in VALID_ICON_FORMATS:
                raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS}, got {ext!r}.')
            prefix = ''
        else:
            if ext not in VALID_ICON_FORMATS_EXTENDED:
                raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS_EXTENDED}, got {ext!r}.')
            prefix = 'a_'
    
    return f'{CDN_ENDPOINT}/guilds/{guild.id}/users/{user.id}/avatars/{prefix}{guild_profile.avatar_hash:0>32x}.' \
           f'{ext}{end}'


def user_avatar_url_at(user, guild):
    """
    Returns the user's avatar's url at the guild.
    
    Parameters
    ----------
    user : ``UserBase``
        The Respective user.
    guild : ``Guild``
        The respective guild.
    
    Returns
    -------
    url : `None` or `str`
    """
    avatar_url = user_avatar_url_for(user, guild)
    if avatar_url is None:
        avatar_url = user_avatar_url(user)
    
    return avatar_url


def user_avatar_url_at_as(user, guild, ext=None, size=None):
    """
    Returns the user's avatar's url at the guild. If the user has no avatar, then returns it's default avatar's url.
    
    Parameters
    ----------
    user : ``UserBase``
        The Respective user.
    guild : ``Guild``
        The respective guild.
    ext : `str`, Optional
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`. If the user has
        animated avatar, it can `'gif'` as well.
    size : `int`, Optional
        The preferred minimal size of the avatar's url.
    
    Returns
    -------
    url : `None` or `str`
    
    Raises
    ------
    ValueError
        If `ext` or `size` was not passed as any of the expected values.
    """
    avatar_url = user_avatar_url_for_as(user, guild, ext=ext, size=size)
    if avatar_url is None:
        avatar_url = user_avatar_url_as(user, ext=ext, size=size)
    
    return avatar_url


def default_avatar_url(default_avatar):
    """
    Returns the user's default avatar's url.
    
    Parameters
    ----------
    user : ``UserBase``
        The Respective user.
    
    Returns
    -------
    url : `str`
    """
    return f'{CDN_ENDPOINT}/embed/avatars/{default_avatar.value}.png'


def application_icon_url(application):
    """
    Returns the application's icon's url. If the application has no icon, then returns `None`.
    
    Parameters
    ----------
    application : ``Application``, ``MessageApplication``, ``IntegrationApplication``
        The respective application.
    
    Returns
    -------
    url : `None` or `str`
    """
    icon_type = application.icon_type
    if icon_type is ICON_TYPE_NONE:
        return None
    
    if icon_type is ICON_TYPE_STATIC:
        prefix = ''
        ext = 'png'
    else:
        prefix = 'a_'
        ext = 'gif'
    
    return f'{CDN_ENDPOINT}/app-icons/{application.id}/{prefix}{application.icon_hash:0>32x}.{ext}'


def application_icon_url_as(application, ext=None, size=None):
    """
    Returns the application's icon's url. If the application has no icon, then returns `None`.
    
    Parameters
    ----------
    application : ``Application``, ``MessageApplication``, ``IntegrationApplication``
        The respective application.
    ext : `str`, Optional
        The extension of the icon's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
    size : `int`, Optional
        The preferred minimal size of the icon's url.
    
    Returns
    -------
    url : `None` or `str`
    
    Raises
    ------
    ValueError
        If `ext` or `size` was not passed as any of the expected values.
    """
    icon_type = application.icon_type
    if icon_type is ICON_TYPE_NONE:
        return None
    
    if size is None:
        end = ''
    elif size in VALID_ICON_SIZES:
        end = f'?size={size}'
    else:
        raise ValueError(f'Size must be in {sorted(VALID_ICON_SIZES)!r}, got {size}.')
    
    if ext is None:
        if icon_type is ICON_TYPE_STATIC:
            prefix = ''
            ext = 'png'
        else:
            prefix = 'a_'
            ext = 'gif'
    
    else:
        if icon_type is ICON_TYPE_STATIC:
            if ext not in VALID_ICON_FORMATS:
                raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS}, got {ext!r}.')
            prefix = ''
        else:
            if ext not in VALID_ICON_FORMATS_EXTENDED:
                raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS_EXTENDED}, got {ext!r}.')
            prefix = 'a_'

    return f'{CDN_ENDPOINT}/app-icons/{application.id}/{prefix}{application.icon_hash:0>32x}.{ext}{end}'


def application_cover_url(application):
    """
    Returns the application's cover image's url. If the application has no cover image, then returns `None`.
    
    Parameters
    ----------
    application : ``Application``, ``MessageApplication``
        The respective application.
        
    Returns
    -------
    url : `None` or `str`
    """
    icon_type = application.cover_type
    if icon_type is ICON_TYPE_NONE:
        return None
    
    if icon_type is ICON_TYPE_STATIC:
        prefix = ''
        ext = 'png'
    else:
        prefix = 'a_'
        ext = 'gif'
    
    return f'{CDN_ENDPOINT}/app-assets/{application.id}/store/{prefix}{application.cover_hash:0>32x}.{ext}'


def application_cover_url_as(application, ext=None, size=None):
    """
    Returns the application's cover image's url. If the application has no cover image, then returns `None`.
    
    Parameters
    ----------
    application : ``Application``, ``MessageApplication``
        The respective application.
    ext : `str`, Optional
        The extension of the cover's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
    size : `int`, Optional
        The preferred minimal size of the cover's url.
    
    Returns
    -------
    url : `None` or `str`
    
    Raises
    ------
    ValueError
        If `ext` or `size` was not passed as any of the expected values.
    """
    icon_type = application.cover_type
    if icon_type is ICON_TYPE_NONE:
        return None
    
    if size is None:
        end = ''
    elif size in VALID_ICON_SIZES:
        end = f'?size={size}'
    else:
        raise ValueError(f'Size must be in {sorted(VALID_ICON_SIZES)!r}, got {size}.')
    
    if ext is None:
        if icon_type is ICON_TYPE_STATIC:
            prefix = ''
            ext = 'png'
        else:
            prefix = 'a_'
            ext = 'gif'
    
    else:
        if icon_type is ICON_TYPE_STATIC:
            if ext not in VALID_ICON_FORMATS:
                raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS}, got {ext!r}.')
            prefix = ''
        else:
            if ext not in VALID_ICON_FORMATS_EXTENDED:
                raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS_EXTENDED}, got {ext!r}.')
            prefix = 'a_'
    
    return f'{CDN_ENDPOINT}/app-assets/{application.id}/store/{prefix}{application.cover_hash:0>32x}.{ext}{end}'


def team_icon_url(team):
    """
    Returns the team's icon's url. If the team has no icon, then returns `None`.
    
    Parameters
    ----------
    team : ``Team``
        The respective team.
    
    Returns
    -------
    url : `None` or `str`
    """
    icon_type = team.icon_type
    if icon_type is ICON_TYPE_NONE:
        return None
    
    if icon_type is ICON_TYPE_STATIC:
        prefix = ''
        ext = 'png'
    else:
        prefix = 'a_'
        ext = 'gif'
    
    return f'{CDN_ENDPOINT}/team-icons/{team.id}/{prefix}{team.icon_hash:0>32x}.{ext}'


def team_icon_url_as(team, ext=None, size=None):
    """
    Returns the team's icon's url. If the team has no icon, then returns `None`.
    
    Parameters
    ----------
    team : ``Team``
        The respective team.
    ext : `str`, Optional
        The extension of the icon's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
    size : `int`, Optional
        The preferred minimal size of the icon's url.
    
    Returns
    -------
    url : `None` or `str`
    
    Raises
    ------
    ValueError
        If `ext` or `size` was not passed as any of the expected values.
    """
    icon_type = team.icon_type
    if icon_type is ICON_TYPE_NONE:
        return None
    
    if size is None:
        end = ''
    elif size in VALID_ICON_SIZES:
        end = f'?size={size}'
    else:
        raise ValueError(f'Size must be in {sorted(VALID_ICON_SIZES)!r}, got {size}.')
    
    if ext is None:
        if icon_type is ICON_TYPE_STATIC:
            prefix = ''
            ext = 'png'
        else:
            prefix = 'a_'
            ext = 'gif'
    
    else:
        if icon_type is ICON_TYPE_STATIC:
            if ext not in VALID_ICON_FORMATS:
                raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS}, got {ext!r}.')
            prefix = ''
        else:
            if ext not in VALID_ICON_FORMATS_EXTENDED:
                raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS_EXTENDED}, got {ext!r}.')
            prefix = 'a_'
    
    return f'{CDN_ENDPOINT}/team-icons/{team.id}/{prefix}{team.icon_hash:0>32x}.{ext}{end}'


def achievement_icon_url(achievement):
    """
    Returns the achievement's icon's url.
    
    Parameters
    ----------
    achievement : ``Achievement``
        The respective achievements
    
    Returns
    -------
    url : `None` or `str`
    """
    icon_type = achievement.icon_type
    if icon_type is ICON_TYPE_NONE:
        return None
    
    if icon_type is ICON_TYPE_STATIC:
        prefix = ''
        ext = 'png'
    else:
        prefix = 'a_'
        ext = 'gif'
    
    return f'{CDN_ENDPOINT}/app-assets/{achievement.application_id}/achievements/{achievement.id}/icons/{prefix}' \
           f'{achievement.icon_hash:0>32x}.{ext}'


def achievement_icon_url_as(achievement, ext=None, size=None):
    """
    Returns the achievement's icon's url.
    
    Parameters
    ----------
    achievement : ``Achievement``
        The respective achievements
    ext : `str`, Optional
        The extension of the icon's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
    size : `int`, Optional
        The preferred minimal size of the icon's url.
    
    Returns
    -------
    url : `None` or `str`
    
    Raises
    ------
    ValueError
        If `ext` or `size` was not passed as any of the expected values.
    """
    icon_type = achievement.icon_type
    if icon_type is ICON_TYPE_NONE:
        return None
    
    if size is None:
        end = ''
    elif size in VALID_ICON_SIZES:
        end = f'?size={size}'
    else:
        raise ValueError(f'Size must be in {sorted(VALID_ICON_SIZES)!r}, got {size}.')
    
    if ext is None:
        if icon_type is ICON_TYPE_STATIC:
            prefix = ''
            ext = 'png'
        else:
            prefix = 'a_'
            ext = 'gif'
    
    else:
        if icon_type is ICON_TYPE_STATIC:
            if ext not in VALID_ICON_FORMATS:
                raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS}, got {ext!r}.')
            prefix = ''
        else:
            if ext not in VALID_ICON_FORMATS_EXTENDED:
                raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS_EXTENDED}, got {ext!r}.')
            prefix = 'a_'
    
    return f'{CDN_ENDPOINT}/app-assets/{achievement.application_id}/achievements/{achievement.id}/icons/{prefix}' \
           f'{achievement.icon_hash:0>32x}.{ext}{end}'


def sticker_url(sticker):
    """
    Returns the sticker's url.
    
    Parameters
    ----------
    sticker : ``Sticker``
        The respective sticker.
    
    Returns
    -------
    url : `None` or `str`
    """
    format = sticker.format
    if format is StickerFormat.none:
        return None
    
    return f'{CDN_ENDPOINT}/stickers/{sticker.id}.{format.extension}'


def sticker_url_as(sticker, size=None, preview=False):
    """
    Returns the sticker's url.
    
    Parameters
    ----------
    sticker : ``Sticker``
        The respective sticker.
    size : `int`, Optional
        The preferred minimal size of the icon's url.
    preview : `bool`, Optional
        Whether preview url should be generated.
    
    Returns
    -------
    url : `None` or `str`
    
    Raises
    ------
    ValueError
        If `size` was not passed as any of the expected values.
    """
    format = sticker.format
    if format is StickerFormat.none:
        return None
    
    # Resolve size
    if size is None:
        end = ''
    else:
        if format is StickerFormat.lottie:
            end = ''
        else:
            if size in VALID_ICON_SIZES:
                end = f'?size={size}'
            else:
                raise ValueError(f'Size must be in {sorted(VALID_ICON_SIZES)!r}, got {size}.')
    
    # Resolve preview
    if preview:
        if format is StickerFormat.apng:
            end = f'{end}{"&" if end else "?"}passthrough=false'
    
    return f'{CDN_ENDPOINT}/stickers/{sticker.id}.{format.extension}{end}'


def sticker_pack_banner(sticker_pack):
    """
    Returns the sticker pack banner's url.
    
    Parameters
    ----------
    sticker_pack : `StickerPack``
        The respective sticker pack.
    
    Returns
    -------
    url : `None` or `str`
    """
    return f'{CDN_ENDPOINT}/app-assets/710982414301790216/store/{sticker_pack.banner_id}.png'


def sticker_pack_banner_as(sticker_pack, ext=None, size=None):
    """
    Returns the achievement's icon's url.
    
    Parameters
    ----------
    sticker_pack : `StickerPack``
        The respective sticker pack.
    ext : `str`, Optional
        The extension of the banner's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
    size : `int`, Optional
        The preferred minimal size of the banner's url.
    
    Returns
    -------
    url : `None` or `str`
    
    Raises
    ------
    ValueError
        If `ext` or `size` was not passed as any of the expected values.
    """
    if size is None:
        end = ''
    elif size in VALID_ICON_SIZES:
        end = f'?size={size}'
    else:
        raise ValueError(f'Size must be in {sorted(VALID_ICON_SIZES)!r}, got {size}.')
    
    if ext is None:
        ext = 'png'
    
    else:
        if ext not in VALID_ICON_FORMATS:
            raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS}, got {ext!r}.')
    
    return f'{CDN_ENDPOINT}/app-assets/710982414301790216/store/{sticker_pack.banner_id}.{ext}{end}'


def role_icon_url(role):
    """
    Returns the role's icon's image's url. If the role has no icon, then returns `None`.
    
    Parameters
    ----------
    role : ``Role``
        The respective role.
    
    Returns
    -------
    url : `None` or `str`
    """
    icon_type = role.icon_type
    if icon_type is ICON_TYPE_NONE:
        return None
    
    if icon_type is ICON_TYPE_STATIC:
        prefix = ''
        ext = 'png'
    else:
        prefix = 'a_'
        ext = 'gif'
    
    return f'{CDN_ENDPOINT}/role-icons/{role.id}/{prefix}{role.icon_hash:0>32x}.{ext}'


def role_icon_url_as(role, ext=None, size=None):
    """
    Returns the role's icon's image's url. If the role has no icon, then returns `None`.
    
    Parameters
    ----------
    role : ``Role``
        The respective role.
    ext : `str`, Optional
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`, `'gif'`.
    size : `int`, Optional
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None` or `str`
    
    Raises
    ------
    ValueError
        If `ext` or `size` was not passed as any of the expected values.
    """
    icon_type = role.icon_type
    if icon_type is ICON_TYPE_NONE:
        return None
    
    if size is None:
        end = ''
    elif size in VALID_ICON_SIZES:
        end = f'?size={size}'
    else:
        raise ValueError(f'Size must be in {sorted(VALID_ICON_SIZES)!r}, got {size}.')
    
    if ext is None:
        if icon_type is ICON_TYPE_STATIC:
            prefix = ''
            ext = 'png'
        else:
            prefix = 'a_'
            ext = 'gif'
    
    else:
        if icon_type is ICON_TYPE_STATIC:
            if ext not in VALID_ICON_FORMATS:
                raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS}, got {ext!r}.')
            prefix = ''
        else:
            if ext not in VALID_ICON_FORMATS_EXTENDED:
                raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS_EXTENDED}, got {ext!r}.')
            prefix = 'a_'

    return f'{CDN_ENDPOINT}/role-icons/{role.id}/{prefix}{role.icon_hash:0>32x}.{ext}{end}'



__all__ = ('UserBase', )

from ...backend.export import include

from ..bases import DiscordEntity, IconSlot, ICON_TYPE_NONE
from ..utils import DATETIME_FORMAT_CODE
from ..color import Color

from ..http import urls as module_urls

from .preinstanced import Status, DefaultAvatar
from .flags import UserFlag
from .helpers import get_banner_color_from_data

create_partial_role_from_id = include('create_partial_role_from_id')
Client = include('Client')


class UserBase(DiscordEntity, immortal=True):
    """
    Base class for user instances.
    
    Attributes
    ----------
    id : `int`
        The client's unique identifier number.
    name : str
        The client's username.
    discriminator : `int`
        The client's discriminator. Given to avoid overlapping names.
    avatar_hash : `int`
        The user's avatar's hash in `uint128`.
    avatar_type : ``IconType``
        The user's avatar's type.
    banner_color : `None` or ``Color``
        The user's banner color if has any.
    banner_hash : `int`
        The user's banner's hash in `uint128`.
    banner_type : ``IconType``
        The user's banner's type.
    
    Notes
    -----
    Instances of this class are weakreferable.
    """
    __slots__ = ('banner_color', 'name', 'discriminator', )
    
    avatar = IconSlot('avatar', 'avatar', module_urls.user_avatar_url, module_urls.user_avatar_url_as)
    banner = IconSlot('banner', 'banner', module_urls.user_banner_url, module_urls.user_banner_url_as)
    
    
    def _update_attributes(self, data):
        """
        Updates the user with the given data by overwriting it's old attributes.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            User data received from Discord.
        """
        try:
            name = data['username']
        except KeyError:
            # Webhook?
            name = data.get('name', None)
            if (name is None):
                name = ''
        
        self.name = name
        
        try:
            discriminator = data['discriminator']
        except KeyError:
            discriminator = 0
        else:
            discriminator = int(discriminator)
        self.discriminator = discriminator
        
        self._set_avatar(data)
        self._set_banner(data)
        
        self.banner_color = get_banner_color_from_data(data)
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the user and returns it's overwritten attributes as a `dict` with a `attribute-name` - `old-value`
        relation.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            User data received from Discord.
        
        Returns
        -------
        old_attributes : `dict` of (`str`, `Any`) items
            All item in the returned dict is optional.
        
        Returned Data Structure
        -----------------------
        
        +---------------+-----------------------+
        | Keys          | Values                |
        +===============+=======================+
        | avatar        | ``Icon``              |
        +---------------+-----------------------+
        | banner        | ``Icon``              |
        +---------------+-----------------------+
        | banner_color  | `None` or ``Color``   |
        +---------------+-----------------------+
        | discriminator | `int`                 |
        +---------------+-----------------------+
        | name          | `str`                 |
        +---------------+-----------------------+
        """
        old_attributes = {}
        
        try:
            name = data['username']
        except KeyError:
            # Webhook?
            name = data.get('name', None)
            if (name is None):
                name = ''
        if self.name != name:
            old_attributes['name'] = self.name
            self.name = name
        
        
        discriminator = int(data['discriminator'])
        if self.discriminator != discriminator:
            old_attributes['discriminator'] = self.discriminator
            self.discriminator = discriminator
        
        
        self._update_avatar(data, old_attributes)
        self._update_banner(data, old_attributes)
        
        
        banner_color = get_banner_color_from_data(data)
        if self.banner_color != banner_color:
            old_attributes['banner_color'] = self.banner_color
            self.banner_color = banner_color
        
        return old_attributes
    
    
    def __repr__(self):
        """Returns the user's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
        ]
        
        repr_parts.append(' id=')
        repr_parts.append(repr(self.id))
        
        if self.partial:
            repr_parts.append(' (partial)')
        else:
            repr_parts.append(', name=')
            repr_parts.append(repr(self.full_name))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def __format__(self, code):
        """
        Formats the user in a format string.
        
        Parameters
        ----------
        code : `str`
            The option on based the result will be formatted.
        
        Returns
        -------
        user : `str`
        
        Raises
        ------
        ValueError
            Unknown format code.
        
        Examples
        --------
        ```py
        >>> from hata import User, now_as_id
        >>> user = User.precreate(now_as_id(), name='Neko', discriminator=2012)
        >>> user
        <User partial, id=730233383967260672>
        >>> # no code stands for `user.name`.
        >>> f'{user}'
        'Neko'
        >>> # 'f' stands for full name
        >>> f'{user:f}'
        'Neko#2012'
        >>> # 'm' stands for mention.
        >>> f'{user:m}'
        '<@730233383967260672>'
        >>> # 'c' stands for created at.
        >>> f'{user:c}'
        '2020.07.08-01:26:45'
        ```
        """
        if not code:
            return self.name
        
        if code == 'f':
            return self.full_name
        
        if code == 'm':
            return self.mention
        
        if code == 'c':
            return self.created_at.__format__(DATETIME_FORMAT_CODE)
        
        raise ValueError(f'Unknown format code {code!r} for object of type {self.__class__.__name__!r}')
    
    
    @property
    def full_name(self):
        """
        The user's name with it's discriminator.
        
        Returns
        -------
        full_name : `str`
        """
        return f'{self.name}#{self.discriminator:0>4}'
    
    
    @property
    def mention(self):
        """
        The mention of the user.
        
        Returns
        -------
        mention : `str`
        """
        return f'<@{self.id}>'
    
    
    @property
    def mention_nick(self):
        """
        The mention to the user's nick.
        
        Returns
        -------
        mention : `str`
        
        Notes
        -----
        It actually has nothing to do with the user's nickname > <.
        """
        return f'<@!{self.id}>'
    
    
    @property
    def default_avatar_url(self):
        """
        Returns the user's default avatar's url.
        
        Returns
        -------
        default_avatar_url : `str`
        """
        return DefaultAvatar.for_(self).url
    
    
    @property
    def default_avatar(self):
        """
        Returns the user's default avatar.
        
        Returns
        -------
        default_avatar : ``DefaultAvatar``
        """
        return DefaultAvatar.for_(self)
    
    
    # for sorting users
    def __gt__(self, other):
        """Returns whether the user's id is greater than the other's."""
        if isinstance(other, UserBase):
            return self.id > other.id
        return NotImplemented
    
    
    def __ge__(self, other):
        """Returns whether the user's id is greater or equal to the other."""
        if isinstance(other, UserBase):
            return self.id >= other.id
        return NotImplemented
    
    
    def __eq__(self, other):
        """Return whether the user's id is equal to the other."""
        if isinstance(other, UserBase):
            return self.id == other.id
        return NotImplemented
    
    
    def __ne__(self, other):
        """Returns whether the user's id is different as the other's."""
        if isinstance(other, UserBase):
            return self.id != other.id
        return NotImplemented
    
    
    def __le__(self, other):
        """Returns whether the user's id is less or equal to the other."""
        if isinstance(other,UserBase):
            return self.id <= other.id
        return NotImplemented
    
    
    def __lt__(self, other):
        """Returns whether the user's id is less than the other's."""
        if isinstance(other, UserBase):
            return self.id < other.id
        return NotImplemented
    
    
    @property
    def activities(self):
        """
        Returns the user's activities.
        
        Returns
        -------
        activities : `None`
        """
        return None
    
    
    @property
    def status(self):
        """
        Returns the user's display status.
        
        Returns
        -------
        status  ``Status``
        """
        return Status.offline
    
    
    @property
    def statuses(self):
        """
        Returns the user's statuses for each platform.
        
        Returns
        -------
        statuses : `dict` of (`str`, `str`) items
        """
        return {}
    
    
    @property
    def guild_profiles(self):
        """
        Returns a dictionary, which contains the user's guild profiles. If the user is member of a guild, then it should
        have a respective guild profile accordingly.
        
        Returns
        -------
        guild_profiles : `dict` of (`int`, ``GuildProfile``) items
        """
        return {}
    
    
    @property
    def thread_profiles(self):
        """
        A Dictionary which contains the thread profiles for the user in thread channel - thread profile relation.
        Defaults to `None`.
        
        Returns
        -------
        thread_profiles : `None` or `dict` (``ChannelThread``, ``ThreadProfile``) items
        """
        return None
    
    
    @property
    def is_bot(self):
        """
        Returns whether the user is a bot or a user account.
        
        Returns
        -------
        is_bot : `bool`
        """
        return False
    
    
    @property
    def flags(self):
        """
        Returns the user's flags.
        
        Returns
        -------
        flags : ``UserFlag``
        """
        return UserFlag()
    
    
    @property
    def partial(self):
        """
        Returns whether the user is partial. Partial users have only their ``.id`` set and every other field might not
        reflect the reality.
        
        Returns
        -------
        partial : `bool`
        """
        return True
    
    
    @property
    def activity(self):
        """
        Returns the user's top activity if applicable. If not.
        
        Returns
        -------
        activity : ``ActivityRich`` or `None`
        """
        return None
    
    
    @property
    def custom_activity(self):
        """
        Returns the user's custom activity if applicable.
        
        Returns
        -------
        activity : ``ActivityCustom`` or `None`
        """
        return None
    
    
    @property
    def platform(self):
        """
        Returns the user's top status's platform. If the user is offline it will return an empty string.
        
        Returns
        -------
        platform : `str`
        """
        return ''
    
    
    def color_at(self, guild):
        """
        Returns the user's color at the given guild.
        
        Parameters
        ----------
        guild : `None` or ``Guild``
            The guild, where the user's color will be checked.
            
            Can be given as `None`.

        Returns
        -------
        color : ``Color``
        """
        return Color()
    
    
    def name_at(self, guild):
        """
        Returns the user's name at the given guild.
        
        Parameters
        ----------
        guild : `None` or ``Guild``
            The guild, where the user's nick will be checked.
            
            Can be given as `None`.

        Returns
        -------
        name : `str`
        """
        return self.name
    
    
    def mentioned_in(self, message):
        """
        Returns whether the user is mentioned at a given message.
        
        Parameters
        ----------
        message : ``Message``
            The message, what's mentions will be checked.
        
        Returns
        -------
        mentioned : `bool`
        """
        if message.everyone_mention:
            return True
        
        user_mentions = message.user_mentions
        if (user_mentions is not None) and (self in user_mentions):
            return True
        
        role_mentions = message.role_mentions
        if (role_mentions is not None):
            guild = message.guild
            if (guild is not None):
                try:
                    guild_profile = self.guild_profiles[guild.id]
                except KeyError:
                    pass
                else:
                    roles = guild_profile.roles
                    if (roles is not None):
                        for role in roles:
                            if role in role_mentions:
                                return True
        
        return False
    
    
    def has_role(self, role):
        """
        Returns whether the user has the given role.
        
        Parameters
        ----------
        role : ``Role``
            The role what will be checked.

        Returns
        -------
        has_role : `bool`
        """
        return False
    
    
    def top_role_at(self, guild, default=None):
        """
        Returns the top role of the user at the given guild.
        
        Parameters
        ----------
        guild : ``Guild`` or `None`
            The guild where the user's top role will be looked up.
            
            Can be given as `None`.
        default : `Any`
            If the user is not a member of the guild, or if has no roles there, then the given default value is returned.
            Defaults to `None`.
        
        Returns
        -------
        top_role ``Role`` or `default`
        """
        return default
    
    
    def can_use_emoji(self, emoji):
        """
        Returns whether the user can use the given emoji.
        
        Parameters
        ----------
        emoji : ``Emoji``
            The emoji to check.
        
        Returns
        -------
        can_use_emoji : `bool`
        """
        if emoji.is_unicode_emoji():
            return True
        
        return False
    
    
    def has_higher_role_than(self, role):
        """
        Returns whether the user has higher role than the given role at it's respective guild.
        
        If the user is the owner of the guild, then returns `True` even if it has lower role.
        
        Parameters
        ----------
        role : ``Role``
            The role to check.
        
        Returns
        -------
        has_higher_role_than : `bool`
        """
        return False
    
    
    def has_higher_role_than_at(self, user, guild):
        """
        Returns whether the user has higher role as the other one at the given guild.
        
        Parameters
        ----------
        user : ``User``
            The other user to check.
        guild : ``Guild`` or `None`
            The guild where the users' top roles will be checked.
            
            Can be given as `None`.
        
        Returns
        -------
        has_higher_role_than_at : `bool`
        """
        return False
    
    
    def get_guild_profile_for(self, guild):
        """
        Returns the user's guild profile for the given guild.
        
        Parameters
        ----------
        guild : ``Guild``
            The guild to get guild profile for.
        
        Returns
        -------
        guild_profile : `None` or ``Guild``
        """
        return None
    
    
    def iter_guild_profiles(self):
        """
        Iterates over the guild profiles of the user.
        
        This method is an iterable generator.
        
        Yields
        ------
        guild : ``Guild``
            The guild profile's guild.
        guild_profile : ``GuildProfile``
            The user's guild profile in the guild.
        """
        return
        yield
    
    
    avatar_url_for = property(module_urls.user_avatar_url_for)
    avatar_url_for_as = module_urls.user_avatar_url_for_as
    avatar_url_at = module_urls.user_avatar_url_at
    avatar_url_at_as = module_urls.user_avatar_url_at_as
    
    
    @classmethod
    def _create_empty(cls, user_id):
        self = object.__new__(cls)
        self.id = user_id
        self._set_default_attributes()
        return self
    
    
    def _set_default_attributes(self):
        """
        Sets the user's attribute's to their default.
        """
        self.name = ''
        self.discriminator = 0
        self.avatar_hash = 0
        self.avatar_type = ICON_TYPE_NONE
        self.banner_color = None
        self.banner_hash = 0
        self.banner_type = ICON_TYPE_NONE



__all__ = ('FlagBase', 'ReverseFlagBase', )

class FlagGetDescriptor:
    """
    Returns the flag descriptor's owner's value at a specific bitwise position.
    
    Attributes
    ----------
    shift : `int`
        The bitwise position of the attribute what this flag represents.
    """
    __slots__ = ('shift', )
    def __init__(self, shift):
        self.shift = shift
    
    def __get__(self, instance, type_):
        if instance is None:
            return self
        else:
            return (instance>>self.shift)&1
    
    def __call__(self, value):
        return (value>>self.shift)&1
    
    def __set__(self, obj, value):
        raise AttributeError('can\'t set attribute')
    
    def __delete__(self, obj):
        raise AttributeError('can\'t delete attribute')


class ReverseFlagGetDescriptor(FlagGetDescriptor):
    """
    Returns the flag descriptor's owner's reversed value at a specific bitwise position.
    
    This type is a reversed version of ``FlagGetDescriptor``, so it returns `0` when the value has `1` at the specific
    bitwise position.
    
    Attributes
    ----------
    shift : `int`
        The bitwise position of the attribute, what this flag represents.
    """
    def __get__(self, instance, type_):
        if instance is None:
            return self
        else:
            return ((instance>>self.shift)&1)^1
    
    def __call__(self, value):
        return ((value>>self.shift)&1)^1


class FlagEnabler:
    """
    Enables a specific bitwise flag of a given value by returning a new one with the given bitwise flag enabled.
    
    This type is instanced by ``FlagEnableDescriptor`` objects, when they are accessed as instance attribute.
    
    Attributes
    ----------
    instance : ``FlagMeta`` instance's instance
        The source value, what will be modified.
    shift : `int`
        The bitwise position, what will be modified.
    
    Notes
    -----
    This type is used for disabling specific bitwise values when the source flag is a reversed flag.
    """
    __slots__ = ('instance', 'shift')
    
    def __call__(self):
        instance = self.instance
        return int.__new__(type(instance), (instance|(1<<self.shift)))


class FlagEnableDescriptor(FlagGetDescriptor):
    """
    Descriptor for enabling a specific bitwise value of a flag. After this descriptor is accessed as an instance
    attribute, a ``FlagEnabler`` is returned, and calling that will return a new instance with it's bitwise flag
    value enabled.
    
    Attributes
    ----------
    shift : `int`
        The bitwise position of the attribute what this flag represents.
    
    Notes
    -----
    This type is used for disabling specific bitwise values when the source flag is a reversed flag.
    """
    def __get__(self, instance, type_):
        if instance is None:
            return self
        else:
            result = FlagEnabler()
            result.instance = instance
            result.shift = self.shift
            return result


class FlagDisabler:
    """
    Disables a specific bitwise flag of a given value by returning a new one with the given bitwise flag disabled.
    
    This type is instanced by ``FlagDisableDescriptor`` objects, when they are accessed as instance attribute.
    
    Attributes
    ----------
    instance : ``FlagMeta`` instance's instance
        The source value, what will be modified.
    shift : `int`
        The bitwise position, what will be modified.
    
    Notes
    -----
    This type is used for enable specific bitwise values when the source flag is a reversed flag.
    """
    __slots__ = ('instance', 'shift')
    
    def __call__(self):
        instance = self.instance
        shift = self.shift
        if (instance>>shift)&1:
            return int.__new__(type(instance), (instance^(1<<shift)))
        else:
            return instance


class FlagDisableDescriptor(FlagGetDescriptor):
    """
    Descriptor for disabling a specific bitwise value of a flag. After this descriptor is accessed as an instance
    attribute, a ``FlagDisabler`` is returned, and calling that will return a new instance with it's bitwise flag
    value disabled.
    
    Attributes
    ----------
    shift : `int`
        The bitwise position of the attribute what this flag represents.
    
    Notes
    -----
    This type is used for enabling specific bitwise values when the source flag is a reversed flag.
    """
    def __get__(self, instance, type_):
        if instance is None:
            return self
        else:
            result = FlagDisabler()
            result.instance = instance
            result.shift = self.shift
            return result


class FlagMeta(type):
    """
    Metaclass for bitwise flags.
    """
    def __new__(cls, class_name, class_parents, class_attributes, access_keyword=None, enable_keyword=None,
            disable_keyword=None, base_class=False):
        """
        Creates a bitwise flag type.
        
        Parameters
        ----------
        class_name : `str`
            The created class's name.
        class_parents : `tuple` of `type` instances
            The superclasses of the creates type.
        class_attributes : `dict` of (`str`, `Any`) items
            The class attributes of the created type.
        access_keyword : `str`, Optional
            The string what will go before the descriptor's name, which allow accessing their respective value. If not
            defined, the there will be no keyword before them.
        enable_keyword : `str`, Optional
            Whether enabling descriptors should be added to the created type. The passed string will be before the
            enabling descriptor's name. If not defined, then the created type will not have flag enablers.
        disable_keyword : `str`, Optional
            Whether disabling descriptors should be added to the created type. The passed string will be before the
            disabling descriptor's name. If not defined, then the created type will not have flag disablers.
        base_class : `bool`, Optional
            Whether the created type is a base class for flags. If it is:
                - Should directly derived from `int`.
                - Should not implement `__keys__` (except `NotImplemented`).
                - Should implement `__getter_class__` class attribute.
                - Should implement `__enabler_class__` class attribute.
                - Should implement `__disabler_class__` class attribute.
            If not, then:
                - Should derive directly from a ``FlagMeta`` base_class.
                - Should implement `__keys__` as `dict` of (`str`, `int`) items, where the values are not less than
                    `0`, and not greater than `63` either.
        
        Returns
        -------
        type : ``FlagMeta`` instance
        
        Raises
        ------
        TypeError
            When any requirements are not satisfied.
        """
        if base_class:
            if (not class_parents) or (not issubclass(class_parents[0], int)):
                raise TypeError(f'`{class_name}` is not derived directly from `int`.')
            
            class_keys = class_attributes.get('__keys__', ...)
            if class_keys is NotImplemented:
                pass
            elif class_keys is ...:
                class_attributes['__keys__'] = NotImplemented
            else:
                raise TypeError(f'`{class_name}` has `__keys__` defined and not as `NotImplemented`.')
            
            for attribute_name in ('__getter_class__', '__enabler_class__', '__disabler_class__'):
                if attribute_name not in class_attributes:
                    raise TypeError(f'`{class_name}` should implement a `{attribute_name}`.')
            
            # do not care about the leftover
            
            return type.__new__(cls, class_name, class_parents, class_attributes)
        
        # Python has no GOTO, so lets insert one
        while True:
            if class_parents:
                parent = class_parents[0]
                if (type(parent) is FlagMeta) and parent.__keys__ is NotImplemented:
                    break
            
            raise TypeError(f'`{class_name}` is not derived directly from a `{cls.__name__}` base instance.')
        
        # Validate keys
        try:
            keys = class_attributes['__keys__']
        except KeyError:
            raise TypeError(f'`{class_name}` did not define `__keys__` attribute.') from None
        
        if (type(keys) is not dict):
            raise TypeError(f'`__keys__` defined as non `dict`: `{keys.__class__.__name__}`.')
        
        for name, shift in keys.items():
            if (type(name) is not str):
                raise TypeError('`__keys__`\'s keys should be `str` instances, meanwhile got at least 1 non `str`: '
                    f'{name!r}.')
            
            if (type(shift) is not int):
                raise TypeError('`__keys__`\'s values should be `int` instances, meanwhile got at least 1 non `int`: '
                    f'{shift!r}.')
            
            if shift < 0 or shift > 63:
                raise TypeError(f'`__keys__`\' values must be between 0 and 63, got: {shift!r}')
        
        class_attributes.setdefault('__new__', int.__new__)
        
        getter = parent.__getter_class__
        enabler = parent.__enabler_class__
        disabler = parent.__disabler_class__
        
        # Add properties
        for name, shift in keys.items():
            if access_keyword is None:
                access_name = name
            else:
                access_name = f'{access_keyword}_{name}'
            
            class_attributes[access_name] = getter(shift)
            
            if (enable_keyword is not None):
                class_attributes[f'{enable_keyword}_{name}'] = enabler(shift)
            
            if (disable_keyword is not None):
                class_attributes[f'{disable_keyword}_{name}'] = disabler(shift)
        
        return type.__new__(cls, class_name, class_parents, class_attributes)


class FlagBase(int, metaclass = FlagMeta, base_class=True):
    """
    Base class for bitwise flags.
    
    Class Attributes
    ----------------
    __getter_class__ : ``FlagGetDescriptor``
        Flag value getter descriptor for subclasses.
    __enabler_class__ : ``FlagEnableDescriptor``
        Flag enabler descriptor for subclasses.
    __disabler_class__ : ``FlagDisableDescriptor``
        Flag disabler descriptor for subclasses.
    """
    __slots__ = ()
    __keys__ = NotImplemented
    
    __getter_class__ = FlagGetDescriptor
    __enabler_class__ = FlagEnableDescriptor
    __disabler_class__ = FlagDisableDescriptor
    
    def __new__(self, base=None):
        """You cannot subclass flag base classes."""
        raise NotImplementedError
    
    def __repr__(self):
        """Returns the representation of the flag."""
        return f'{self.__class__.__name__}({int.__repr__(self)})'
    
    def __getitem__(self, key):
        """Returns whether a specific flag of the given name is enabled."""
        return (self>>self.__keys__[key])&1
    
    def keys(self):
        """
        Yields the name of the bitwise flags, which are enabled.
        
        This method is a generator.
        
        Yields
        ------
        name : `str`
        """
        for name, shift in self.__keys__.items():
            if (self>>shift)&1:
                yield name
    
    __iter__ = keys
    
    def values(self):
        """
        Yields the shift values of the flags, under which shift value the flag is enabled.
        
        This method is a generator.
        
        Yields
        -------
        shift : `int`
        """
        for shift in self.__keys__.values():
            if (self>>shift)&1:
                yield shift
    
    def items(self):
        """
        Yields the items of the flag.
        
        This method is a generator.
        
        Yields
        ------
        name : `str`
            The name of the specific flag
        enabled : `int` (`0` or `1`)
            Whether the specific bitwise value is enabled.
        """
        for name, shift in self.__keys__.items():
            yield name, (self>>shift)&1
    
    def __contains__(self, key):
        """Returns whether the specific flag of the given name is enabled."""
        try:
            position = self.__keys__[key]
        except KeyError:
            return 0
        
        return (self>>position)&1
    
    def is_subset(self, other):
        """Returns whether self has the same amount or more flags disabled than other."""
        return (self&other) == self
    
    def is_superset(self, other):
        """Returns whether self has the same amount or more flags enabled than other."""
        return (self|other) == self
    
    def is_strict_subset(self, other):
        """Returns whether self has more flags disabled than other."""
        return self != other and (self&other) == self
    
    def is_strict_superset(self, other):
        """Returns whether self has more flags enabled than other."""
        return self != other and (self|other) == self
    
    __ge__ = is_superset
    __gt__ = is_strict_superset
    __lt__ = is_strict_subset
    __le__ = is_subset
    
    def update_by_keys(self, **kwargs):
        """
        Updates the source value with the given flags and returns a new one.
        
        Parameters
        ----------
        **kwargs : Keyword parameters
            `flag-name` - `bool` relations.
        
        Returns
        -------
        flag : ``FlagBase`` instance
        
        Raises
        ------
        LookupError
            If a keyword is invalid.
        
        Examples
        -------
        ```py
        >>> from hata import Permission
        >>> perm = Permission().update_by_keys(kick_users=True, ban_users=True)
        >>> list(perm)
        ['kick_users', 'ban_users']
        >>> perm = perm.update_by_keys(manage_roles=True, kick_users=False)
        >>> list(perm)
        ['ban_users', 'manage_roles']
        ```
        """
        new = self
        for key, value in kwargs.items():
            try:
                shift = self.__keys__[key]
            except KeyError:
                raise LookupError(f'Invalid key: {key!r}.') from None
            
            if value:
                new |= (1<<shift)
            else:
                if (new>>shift)&1:
                    new ^= (1<<shift)
        
        return int.__new__(type(self), new)

class ReverseFlagBase(FlagBase, base_class=True):
    """
    Base class for reversed bitwise flags.
    
    Class Attributes
    ----------------
    __getter_class__ : ``ReverseFlagGetDescriptor``
        Flag value getter descriptor for subclasses.
    __enabler_class__ : ``FlagDisableDescriptor``
        Flag enabler descriptor for subclasses.
    __disabler_class__ : ``FlagEnableDescriptor``
        Flag disabler descriptor for subclasses.
    """
    __getter_class__ = ReverseFlagGetDescriptor
    __enabler_class__ = FlagDisableDescriptor
    __disabler_class__ = FlagEnableDescriptor
    
    def __getitem__(self, key):
        """Returns whether a specific flag of the given name is enabled."""
        return ((self>>self.__keys__[key])&1)^1
    
    def keys(self):
        """
        Yields the name of the bitwise flags, which are enabled.
        
        This method is a generator.
        
        Yields
        ------
        name : `str`
        """
        for name, shift in self.__keys__.items():
            if ((self>>shift)&1)^1:
                yield name
    
    __iter__ = keys
    
    def values(self):
        """
        Yields the shift values of the flags, under which shift value the flag is enabled.
        
        This method is a generator.
        
        Yields
        -------
        shift : `int`
        """
        for shift in self.__keys__.values():
            if ((self>>shift)&1)^1:
                yield shift
    
    def items(self):
        """
        Yields the items of the flag.
        
        This method is a generator.
        
        Yields
        -------
        name : `str`
            The name of the specific flag
        enabled : `int` (`0` or `1`)
            Whether the specific bitwise value is enabled.
        """
        for name, shift in self.__keys__.items():
            yield name, ((self>>shift)&1)^1
    
    def __contains__(self, key):
        """Returns whether the specific flag of the given name is enabled."""
        try:
            position = self.__keys__[key]
        except KeyError:
            return 0
        
        return ((self>>position)&1)^1
    
    def is_subset(self, other):
        """Returns whether self has the same amount or more flags disabled than other."""
        return (self|other) == self
    
    def is_superset(self, other):
        """Returns whether self has the same amount or more flags enabled than other."""
        return (self&other) == self
    
    def is_strict_subset(self, other):
        """Returns whether self has more flags disabled than other."""
        return self != other and (self|other) == self
    
    def is_strict_superset(self, other):
        """Returns whether self has more flags enabled than other."""
        return self != other and (self&other) == self
    
    __ge__ = is_superset
    __gt__ = is_strict_superset
    __lt__ = is_strict_subset
    __le__ = is_subset
    
    def update_by_keys(self, **kwargs):
        """
        Updates the source value with the given flags and returns a new one.
        
        Parameters
        ----------
        **kwargs : Keyword parameters
            `flag-name` - `bool` relations.
        
        Returns
        -------
        flag : ``ReverseFlagBase`` instance
        
        Examples
        --------
        ```py
        >>> from hata import SystemChannelFlag
        >>> flags = SystemChannelFlag()
        >>> list(flags)
        ['welcome', 'boost']
        >>> flags = flags.update_by_keys(boost=False)
        >>> list(flags)
        ['welcome']
        ```
        """
        new = self
        for key, value in kwargs.items():
            try:
                shift = self.__keys__[key]
            except KeyError as err:
                err.args = (f'Invalid key:{key!r}.',)
                raise
            
            if value:
                if (new>>shift)&1:
                    new ^= (1<<shift)
            else:
                new |= (1<<shift)
        
        return int.__new__(type(self), new)



__all__ = ('WebhookType',)

from ..bases import PreinstancedBase, Preinstance as P

class WebhookType(PreinstancedBase):
    """
    Represents a webhook's type.
    
    Attributes
    ----------
    name : `str`
        The name of the webhook type.
    value : `int`
        The discord side identifier value of the webhook type.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``WebhookType``) items
        Stores the predefined ``WebhookType`` instances. These can be accessed with their `value` as key.
    VALUE_TYPE : `type` = `int`
        The webhook types' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the webhook types.
    
    Every predefined webhook type can be accessed as class attribute as well:
    
    +-----------------------+---------------+-------+
    | Class attribute name  | name          | value |
    +=======================+===============+=======+
    | none                  | none          | 0     |
    +-----------------------+---------------+-------+
    | bot                   | bot           | 1     |
    +-----------------------+---------------+-------+
    | server                | server        | 2     |
    +-----------------------+---------------+-------+
    | application           | application   | 3     |
    +-----------------------+---------------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    
    __slots__ = ()
    
    # predefined
    none = P(0, 'none')
    bot = P(1, 'bot')
    server = P(2, 'server')
    application = P(3, 'application')




__all__ = ('reaction_mapping', 'reaction_mapping_line',)

from ...backend.export import include
from ...backend.utils import set_docs

from .utils import create_partial_emoji_from_data

Client = include('Client')


class reaction_mapping(dict):
    """
    A `dict` subclass, which contains the reactions on a ``Message`` with (``Emoji``, ``reaction_mapping_line``)
    items.
    
    Attributes
    ----------
    fully_loaded : `bool`
        Whether the reaction mapping line is fully loaded.
    """
    __slots__ = ('fully_loaded',)
    
    def __init__(self, data):
        """
        Fills the reaction mapping with the given data.
        
        Parameters
        ----------
        data : `None` or `dict` of (`str`, `Any`) items
        """
        if (data is None) or (not data):
            self.fully_loaded = True
            return
        
        self.fully_loaded = False
        for line in data:
            self[create_partial_emoji_from_data(line['emoji'])] = reaction_mapping_line(line.get('count', 1))
    
    emoji_count = set_docs(
        property(dict.__len__),
        """
        The amount of different emojis, which were added on the reaction mapping's respective ``Message``.
        
        Returns
        -------
        emoji_count : `int`
        """
    )
    
    @property
    def total_count(self):
        """
        The total amount reactions given on the reaction mapping's respective message.
        
        Returns
        -------
        total_count : `int`
        """
        count = 0
        for line in self.values():
            count += set.__len__(line)
            count += line.unknown
        return count
    
    
    def clear(self):
        """
        Clears the reaction mapping with clearing it's lines.
        """
        for value in self.values():
            value.clear()
        if self.fully_loaded:
            self._full_check()
    
    
    def add(self, emoji, user):
        """
        Adds a user to the reactors.
        
        Parameters
        ----------
        emoji : ``Emoji``
            The reacted emoji.
        user : ``ClientUserBase``
            The reactor user.
        """
        try:
            line = self[emoji]
        except KeyError:
            line = reaction_mapping_line(0)
            self[emoji] = line
        line.add(user)
    
    
    def remove(self, emoji, user):
        """
        Removes a user to the reactors.
        
        Parameters
        ----------
        emoji : ``Emoji``
            The removed reacted emoji.
        user : ``ClientUserBase``
            The removed reactor user.
        """
        try:
            line = self[emoji]
        except KeyError:
            return

        if set.__len__(line):
            try:
                line.remove(user)
            except KeyError:
                pass
            else:
                if set.__len__(line) or line.unknown:
                    return
                del self[emoji]
                return
        
        if line.unknown:
            line.unknown -=1
            if set.__len__(line):
                if line.unknown:
                    return
                self._full_check()
                return
            if line.unknown:
                return
            del self[emoji]
    
    
    def remove_emoji(self, emoji):
        """
        Removes all the users who reacted with the given ``Emoji`` and then returns the stored line.
        
        Parameters
        ----------
        emoji : ``Emoji``
            The emoji to remove.
        
        Returns
        -------
        line : `None` or ``reaction_mapping_line``
        """
        line = self.pop(emoji, None)
        if line is None:
            return
        
        if line.unknown:
            self._full_check()
        
        return line
    
    
    # this function is called if an emoji loses all it's unknown reactors
    def _full_check(self):
        """
        Checks whether the reaction mapping is fully loaded, by checking it's values' `.unknown` and sets the current
        state to `.fully_loaded`.
        """
        for line in self.values():
            if line.unknown:
                self.fully_loaded = False
                return
        
        self.fully_loaded = True
    
    
    # we call this when we get SOME reactors of an emoji
    def _update_some_users(self, emoji, users):
        """
        Called when some reactors of an emoji are updated.
        
        Parameters
        ----------
        emoji : ``Emoji``
            The emoji, which's users' are updated.
        users : `list` of ``ClientUserBase``
            The added reactors.
        """
        self[emoji].update(users)
        self._full_check()
        
    def _update_all_users(self, emoji, users):
        """
        Called when all the reactors of an emoji are updated of the reaction mapping.
        
        Parameters
        ----------
        emoji : ``Emoji``
            The emoji, which's users' are updated.
        users : `list` of ``ClientUserBase``
            The added reactors.
        """
        self[emoji] = reaction_mapping_line._full(users)
        self._full_check()


class reaction_mapping_line(set):
    """
    A `set` subclass which contains the users who reacted with the given ``Emoji`` on a ``Message``.
    
    Attributes
    ----------
    unknown : `int`
        The amount of not known reactors.
    """
    __slots__ = ('unknown',)
    
    def __init__(self, unknown):
        """
        Creates a `reaction_mapping_line`.
        
        Parameters
        ----------
        unknown : `int`
            The amount of not known reactors.
        """
        self.unknown = unknown
    
    def __len__(self):
        """Returns the amount of users, who reacted with the given emoji on the respective message."""
        return set.__len__(self)+self.unknown
    
    
    def __repr__(self):
        """Returns the representation of the container."""
        repr_parts = [
            self.__class__.__name__,
            '({',
        ]
        
        # set indexing is not public, so we need to do a check, like this
        if set.__len__(self):
            for user in self:
                repr_parts.append(repr(user))
                repr_parts.append(', ')
            
            repr_parts[-1] = '}'
        else:
            repr_parts.append('}')
        
        unknown = self.unknown
        if unknown:
            repr_parts.append(', unknown=')
            repr_parts.append(repr(unknown))
        
        repr_parts.append(')')
        
        return ''.join(repr_parts)
    
    
    @classmethod
    def _full(cls, users):
        """
        Creates a new ``reaction_mapping_line`` with the given users with `.unknown` set to `0`.
        
        Parameters
        ----------
        users : `list` of ``ClientUserBase``
            A `list`, which should already contain all the users of the reaction mapping line.

        Returns
        -------
        self : ``reaction_mapping_line``
        """
        self = set.__new__(cls)
        set.__init__(self, users)
        self.unknown = 0
        return self
    
    
    def update(self, users):
        """
        Updates the reaction mapping line with the given users.
        
        Parameters
        ----------
        users : `list` of ``ClientUserBase``
            A `list` of users, who reacted on the respective `Message` with the respective ``Emoji``.
        """
        ln_old = len(self)
        set.update(self,users)
        ln_new = len(self)
        self.unknown -= (ln_new-ln_old)
    
    
    def copy(self):
        """
        Copies the reaction mapping line.
        
        Returns
        -------
        new : ``reaction_mapping_line``
        """
        new = set.__new__(type(self))
        set.__init__(new,self)
        new.unknown = self.unknown
        return new
    
    
    # executes an api request if we know we know all reactors
    def filter_after(self, limit, after):
        """
        If we know all the reactors, then instead of executing a Discord API request we filter the reactors locally
        using this method.
        
        Parameters
        ----------
        limit : `int`
            The maximal limit of the users to return.
        after : `int`
            Gets the users after this specified id.
        
        Returns
        -------
        users : `
        """
        list_form = sorted(self)
        
        after = after+1 # do not include the specified id
        
        bot = 0
        top = len(list_form)
        while True:
            if bot < top:
                half = (bot+top)>>1
                if list_form[half].id < after:
                    bot = half+1
                else:
                    top = half
                continue
            break
        
        index = bot
        
        length = len(list_form)
        users = []
        
        while True:
            if index == length:
                break
            
            if limit <= 0:
                break
            
            users.append(list_form[index])
            index += 1
            limit -= 1
            continue
        
        return users
    
    
    def clear(self):
        """
        Clears the reaction mapping line by removing every ``User`` object from it.
        """
        clients = []
        for user in self:
            if isinstance(user, Client):
                clients.append(user)

        self.unknown += (set.__len__(self) - len(clients))
        set.clear(self)
        set.update(self,clients)



__all__ = ('get_channel_stdin', 'get_channel_stdout',)

from collections import deque
from datetime import datetime, timedelta

from ...backend.futures import Task, sleep, shield, future_or_timeout, Future, CancelledError
from ...backend.event_loop import LOOP_TIME
from ...discord.core import KOKORO
from ...discord.utils import sanitize_content

MESSAGE_EDIT_TIMEDELTA = timedelta(seconds=10)
REQUEST_RATE_LIMIT = 1.2


class ChannelOutputStream:
    """
    Attributes
    ----------
    _channel : ``ChannelText``
        The target channel of the stream.
    _chunk_size : `int`
        The maximal chunk size to send in one message.
    _client : ``Client``
        The client who send the stream to the channel.
    _close_waiter : `None` or ``Future``
        Waiter future set to wait for closing.
    _closed : `bool`
        Whether the stream is closed.
    _chunks : `str`
        The data queue.
    _last_chunk : `None` or `str`
        The last send raw chunk.
    _last_message : `None` or ``Message``
        The last sent message.
    _sanitize : `bool`
        Whether the output stream should be sanitized.
    _transfer_task : `None` or ``Task``
        Transfer task set meanwhile data is transferred to ``._channel``.
    """
    __slots__ = ('_channel', '_chunk_size', '_client', '_close_waiter', '_closed', '_chunks', '_last_chunk',
        '_last_message', '_sanitize', '_transfer_task')
    
    def __init__(self, client, channel, chunk_size, sanitize):
        """
        Creates a new ``ChannelOutputStream`` with the given parameters.
        
        Parameters
        ----------
        client : ``Client``
            The client who send the stream to the channel.
        channel : ``ChannelText``
            The target channel of the stream.
        chunk_size : `int`
            The maximal chunk size to send in one message.
        sanitize : `bool`
            Whether the output stream should be sanitized.
        """
        self._client = client
        self._channel = channel
        self._chunks = deque()
        self._transfer_task = None
        self._chunk_size = chunk_size
        self._sanitize = sanitize
        self._last_message = None
        self._last_chunk = None
        self._closed = False
        self._close_waiter = None
    
    def write(self, data):
        """
        Writes the given data to the buffer.
        
        Parameters
        ----------
        data : `str`
            The data to write.
        
        Raises
        ------
        TypeError
            Only string can be written to the output stream.
        ValueError
            I/O operation on closed or on a detached file.
        """
        if self._closed:
            raise ValueError('I/O operation on closed or on a detached file.')
        
        if not isinstance(data, str):
            raise TypeError(f'Only `str` can be written into `{self.__class__.__name__}`, got '
                f'{data.__class__.__name__}')
        
        if not data:
            return
        
        self._chunks.append(data)
        
        transfer_task = self._transfer_task
        if (transfer_task is None):
            self._transfer_task = Task(self._do_transfer(), KOKORO)
    
    
    async def flush(self):
        """
        Force the text in the buffer into the raw stream.
        
        After flushing the newly written content will be forced to new message as well.
        
        This method is a coroutine.
        
        Raises
        ------
        BlockingIOError
            If the raw stream blocks.
        ValueError
            I/O operation on closed or on a detached file.
        """
        transfer_task = self._transfer_task
        if (transfer_task is not None):
            await transfer_task
        
        self._last_message = None
        self._last_chunk = None
    
    
    async def _do_transfer(self):
        """
        Sends the data written to the stream to the respective channel.
        
        This method is a coroutine.
        """
        try:
            client = self._client
            while True:
                try:
                    message = self._last_message
                    if (message is None):
                        un_poll = None
                        should_edit = False
                    else:
                        last_action = message.edited_at
                        if last_action is None:
                            last_action = message.created_at
                        
                        if (last_action+MESSAGE_EDIT_TIMEDELTA > datetime.utcnow()):
                            un_poll = self._last_chunk
                            should_edit = True
                        else:
                            un_poll = None
                            should_edit = False
                    
                    raw_data = self._poll(un_poll)
                    if raw_data is None:
                        break
                    
                    if len(raw_data) < (self._chunk_size >> 1):
                        maybe_update_next = True
                    else:
                        maybe_update_next = False
                    
                    if self._sanitize:
                        data = sanitize_content(raw_data, guild=self._channel.guild)
                    else:
                        data = raw_data
                    
                    request_start = LOOP_TIME()
                    if should_edit:
                        await client.message_edit(message, data)
                    else:
                        message = await client.message_create(self._channel, data)
                    
                    if maybe_update_next:
                        self._last_message = message
                        self._last_chunk = raw_data
                    else:
                        self._last_message = None
                        self._last_chunk = None
                    
                    sleep_time = request_start-LOOP_TIME()+REQUEST_RATE_LIMIT
                    
                    if sleep_time > 0.0:
                        await sleep(sleep_time, KOKORO)
                
                except BaseException as err:
                    self._last_message = None
                    self._last_chunk = None
                    await client.events.error(client, f'{self!r}._do_transfer', err)
        finally:
            self._transfer_task = None
    
    
    def _poll(self, un_poll):
        """
        Polls a chunk of data from the stream.
        
        Parameters
        ----------
        un_poll : `None` or `str`
            Data to un-poll.
        
        Returns
        -------
        data : `None` or `str`
            The data to send if any.
        """
        chunks = self._chunks
        
        if not chunks:
            return None
        
        data_parts = []
        data_parts_size = self._chunk_size
        
        if (un_poll is not None):
            chunks.appendleft(un_poll)
        
        while chunks:
            part = chunks[0]
            data_parts_size -= len(part)
            if data_parts_size > 0:
                data_parts.append(part)
                del chunks[0]
                continue
            
            if data_parts_size == 0:
                data_parts.append(part)
                del chunks[0]
                break
            
            data_parts.append(part[:data_parts_size])
            chunks[0] = part[data_parts_size:]
            break
        
        if data_parts:
            return ''.join(data_parts)
    
    
    def __repr__(self):
        """Returns the stream's representation."""
        repr_parts = ['<', self.__class__.__name__, ' client=', repr(self._client), ', channel=', repr(self._channel)]
        
        chunks_size = 0
        for data in self._chunks:
            chunks_size += len(data)
        
        if chunks_size:
            repr_parts.append(' size=')
            repr_parts.append(repr(chunks_size))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    def close(self):
        """
        Closes the output.
        """
        if self._closed:
            return
        
        self._closed = True
        
        close_waiter = self._close_waiter
        if (close_waiter is not None):
            close_waiter.set_result_if_pending(None)
    
    async def wait_for_close(self, timeout=None):
        """
        Waits till the stream is closed.
        
        Parameters
        ----------
        timeout : `None` or `float`, Optional
            Maximal timeout to wait.
        
        Raises
        ------
        TimeoutError
            Timeout occurred.
        """
        if self._closed:
            return
        
        close_waiter = self._close_waiter
        if (close_waiter is None):
            close_waiter = self._close_waiter = Future(KOKORO)
        
        waiter = shield(close_waiter, KOKORO)
        
        if (timeout is not None):
            future_or_timeout(waiter, timeout)
        
        return waiter
    
    def __enter__(self):
        """Enters the stream returning itself."""
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exists the stream by closing it."""
        self.close()

class ChannelInputStream:
    """
    Attributes
    ----------
    _channel : ``ChannelText``
        The source channel of the stream.
    _check : `None` or `callable`
        Optional check to call to check whether a received message should be passed to the stream.
        
        Should accept the following parameters:
        
        +-----------+---------------+
        | Name      | Type          |
        +===========+===============+
        | message   | ``Message``   |
        +-----------+---------------+
        
        Should return the following values:
        
        +-------------------+---------------+
        | Name              | Type          |
        +===================+===============+
        | should_redirect   | ``bool``      |
        +-------------------+---------------+
    
    _client : ``Client``
        The client who send the stream to the channel.
    _close_waiter : `None` or ``Future``
        Waiter future set to wait for closing.
    _closed : `bool`
        Whether the stream is closed.
    _chunks : `str`
        The data queue.
    _payload_reader : `None` or `generator`
        Payload reader generator, what gets the control back, when data, eof or any exception is received.
    _payload_waiter : `None` of ``Future``
        Payload waiter of the protocol, what's result is set, when the ``._payload_reader`` generator returns.
        
        If cancelled or marked by done or any other methods, the payload reader will not be cancelled.
    """
    __slots__ = ('_channel', '_check', '_client', '_close_waiter', '_closed', '_chunks', '_payload_reader',
        '_payload_waiter')
    
    def __init__(self, client, channel, check):
        """
        Creates a new ``ChannelInputStream`` with the given parameters.
        
        Parameters
        ----------
        client : ``Client``
            The client who send the stream to the channel.
        channel : ``ChannelText``
            The target channel of the stream.
        check : `None` or `callable`
            Optional check to call to check whether a received message should be passed to the stream.
        """
        self._client = client
        self._channel = channel
        self._chunks = deque()
        self._check = check
        self._closed = False
        self._close_waiter = None
        self._payload_reader = None
        self._payload_waiter = None
        
        client.events.message_create.append(channel, self)
    
    
    async def __call__(self, client, message):
        """
        Feeds a message to the input.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who received the respective message.
        message : ``Message``
            the received message.
        """
        check = self._check
        if (check is not None):
            should_redirect = check(message)
            if (not isinstance(should_redirect, int)) or (not should_redirect):
                return
        
        data = message.content
        if not data:
            return
        
        
        chunks = self._chunks
        for data in (data, '\n'):
            payload_reader = self._payload_reader
            if payload_reader is None:
                chunks.append(data)
                continue
            
            try:
                payload_reader.send(data)
            except StopIteration as err:
                args = err.args
                if not args:
                    result = None
                elif len(args) == 1:
                    result = args[0]
                else:
                    result = args
                
                payload_waiter = self._payload_waiter
                self._payload_reader = None
                self._payload_waiter = None
                payload_waiter.set_result_if_pending(result)
            except GeneratorExit as err:
                payload_waiter = self._payload_waiter
                self._payload_reader = None
                self._payload_waiter = None
                exception = CancelledError()
                exception.__cause__ = err
                payload_waiter.set_exception_if_pending(exception)
            except BaseException as err:
                payload_waiter = self._payload_waiter
                self._payload_reader = None
                self._payload_waiter = None
                payload_waiter.set_exception_if_pending(err)
    
    def close(self):
        """
        Closes the input.
        """
        if self._closed:
            return
        
        self._closed = True
        self._client.events.message_create.remove(self._channel, self)
    
        close_waiter = self._close_waiter
        if (close_waiter is not None):
            close_waiter.set_result_if_pending(None)
    
        payload_reader = self._payload_reader
        if payload_reader is None:
            return False
        
        try:
             payload_reader.throw(CancelledError())
        except CancelledError as err:
            new_exception = ConnectionError('Connection closed unexpectedly with EOF.')
            new_exception.__cause__ = err
            payload_waiter = self._payload_waiter
            self._payload_reader = None
            self._payload_waiter = None
            payload_waiter.set_exception_if_pending(new_exception)
        
        except StopIteration as err:
            args = err.args
            if not args:
                result = None
            elif len(args) == 1:
                result = args[0]
            else:
                result = args
            
            payload_waiter = self._payload_waiter
            self._payload_reader = None
            self._payload_waiter = None
            payload_waiter.set_result_if_pending(result)
        except GeneratorExit as err:
            payload_waiter = self._payload_waiter
            self._payload_reader = None
            self._payload_waiter = None
            exception = CancelledError()
            exception.__cause__ = err
            payload_waiter.set_exception_if_pending(exception)
        except BaseException as err:
            payload_waiter = self._payload_waiter
            self._payload_reader = None
            self._payload_waiter = None
            payload_waiter.set_exception_if_pending(err)
    
    
    async def wait_for_close(self, timeout=None):
        """
        Waits till the stream is closed.
        
        Parameters
        ----------
        timeout : `None` or `float`, Optional
            Maximal timeout to wait.
        
        Raises
        ------
        TimeoutError
            Timeout occurred.
        """
        if self._closed:
            return
        
        close_waiter = self._close_waiter
        if (close_waiter is None):
            close_waiter = self._close_waiter = Future(KOKORO)
        
        waiter = shield(close_waiter, KOKORO)
        
        if (timeout is not None):
            future_or_timeout(waiter, timeout)
        
        return waiter
    
    
    def _set_payload_reader(self, payload_reader):
        """
        Sets payload reader to the stream.
        
        Parameters
        ----------
        payload_reader : `generator`
            A generator, what gets control, every time a chunk is received, till it returns or raises.
        
        Returns
        -------
        payload_waiter : ``Future``
            Waiter, to what the result of the `payload_reader` is set.
        
        Raises
        ------
        ValueError
            I/O operation on closed or on a detached file.
        """
        assert self._payload_reader is None, 'Payload reader already set!'
        
        if self._closed and (not self._chunks):
            raise ValueError('I/O operation on closed or on a detached file.')
        
        payload_waiter = Future(KOKORO)
        
        try:
            payload_reader.send(None)
        except StopIteration as err:
            args = err.args
            if not args:
                result = None
            elif len(args) == 1:
                result = args[0]
            else:
                result = args
            
            payload_waiter.set_result_if_pending(result)
        except GeneratorExit as err:
            exception = CancelledError()
            exception.__cause__ = err
            payload_waiter.set_exception_if_pending(exception)
        except BaseException as err:
            payload_waiter.set_exception_if_pending(err)
        
        else:
            self._payload_waiter = payload_waiter
            self._payload_reader = payload_reader
        
        return payload_waiter
    
    
    def _read_all(self):
        """
        Reads all the data from the stream.
        
        This method is a generator.
        
        Returns
        -------
        data : `str`
            The read data.
        """
        chunks = []
        self_chunks = self._chunks
        chunks.extend(self_chunks)
        self_chunks.clear()
        
        while True:
            try:
                chunk = yield
            except CancelledError:
                break
            else:
                chunks.append(chunk)
        
        return ''.join(chunks)
    
    
    def _read_exactly(self, size):
        """
        Reads the exact amount of data form the stream.
        
        This method is a generator.
        
        Returns
        -------
        data : `str`
            The read data.
        """
        if size == 0:
            return b''
        
        chunks = []
        self_chunks = self._chunks
        
        while self_chunks:
            chunk = self_chunks[0]
            chunk_length = len(chunk)
            size -= chunk_length
            if size > 0:
                del self_chunks[0]
                chunks.append(chunk)
                continue
            
            if size == 0:
                del self_chunks[0]
                chunks.append(chunk)
                break
            
            self_chunks[0] = chunk[-size:]
            chunks.append(chunk[:-size])
            break
        
        if (size > 0) and (not self._closed):
            while True:
                try:
                    chunk = yield
                except CancelledError:
                    break
                else:
                    chunk_length = len(chunk)
                    size -= chunk_length
                    if size > 0:
                        chunks.append(chunk)
                        continue
                    
                    if size == 0:
                        chunks.append(chunk)
                        break
                    
                    self_chunks.appendleft(chunk[-size:])
                    chunks.append(chunk[:-size])
                    break
        
        return ''.join(chunks)
    
    
    def _read_line(self):
        """
        Reads a line from the stream.
        
        This method is a generator
        
        Returns
        -------
        data : `str`
            The read data.
        """
        chunks = []
        self_chunks = self._chunks
        
        while self_chunks:
            chunk = self_chunks[0]
            index = chunk.find('\n')
            if index == -1:
                del self_chunks[0]
                chunks.append(chunk)
                continue
            
            index += 1
            if index == len(chunk):
                del self_chunks[0]
                chunks.append(chunk)
                line_break_found = True
                break
            
            self_chunks[0] = chunk[index:]
            chunks.append(chunk[:index])
            line_break_found = True
            break
        else:
            line_break_found = False
        
        if (not line_break_found) and (not self._closed):
            while True:
                try:
                    chunk = yield
                except CancelledError:
                    break
                else:
                    index = chunk.find('\n')
                    if index == -1:
                        chunks.append(chunk)
                        continue
                    
                    index += 1
                    if index == len(chunk):
                        chunks.append(chunk)
                        break
                    
                    self_chunks.appendleft(chunk[index:])
                    chunks.append(chunk[:index])
                    break
        
        return ''.join(chunks)
    
    
    def _read_lines(self, limit):
        """
        Reads a line from the stream.
        
        This method is a generator
        
        Parameters
        ----------
        limit : `int`
            The upper limit if lines to read.
            
            If given as negative `int`, then all the data till eof is read from the stream.
        
        Returns
        -------
        lines : `list` of `str`
            The read lines.
        """
        lines = []
        
        while limit:
            line = yield from self._read_line()
            if not line:
                break
            
            lines.append(line)
            limit -= 1
            continue
        
        return lines
    
    
    async def read(self, size=-1):
        """
        Reads from the underlying stream.
        
        This method is a coroutine.
        
        Parameters
        ----------
        size : `int`
            The upper limit to read from the underlying stream. If eof is received meanwhile reading, the already read
            data is returned.
            
            If given as negative `int`, then all the data till eof is read.
        
        Returns
        -------
        data : `str`
            The read data.
        
        Raises
        ------
        ValueError
            I/O operation on closed or on a detached file.
        """
        return await self._set_payload_reader(self._read_all() if size<0 else self._read_exactly(size))
    
    
    async def read_line(self):
        """
        Reads a line from the underlying stream.
        
        This method is a coroutine.
        
        Returns
        -------
        data : `str`
            The read data.
        
        Raises
        ------
        ValueError
            I/O operation on closed or on a detached file.
        """
        return await self._set_payload_reader(self._read_line())
    
    readline = read_line

    async def read_lines(self, hint=-1):
        """
        Read and return a list of lines from the stream.
        
        This method is a coroutine.
        
        Parameters
        ----------
        hint : `int`, Optional
            The upper limit if lines to read. Defaults to `-1`.
            
            If given as negative `int`, then all the data till eof is read from the stream.
        
        Returns
        -------
        lines : `list` of (`bytes` or `str`)
            The red lines.
        
        Raises
        ------
        ValueError
            I/O operation on closed or on a detached file.
        """
        return await self._set_payload_reader(self._read_lines(hint))
    
    readlines = read_lines
    
    def __enter__(self):
        """Enters the stream returning itself."""
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exists the stream by closing it."""
        self.close()


def get_channel_stdout(client, channel, *, chunk_size=1000, sanitize=False):
    """
    Gets output stream towards the given channel.
    
    Parameters
    ----------
    client : ``Client``
        The client who redirects the stream.
    channel : ``ChannelText``
        The channel where the stream is redirected from.
    chunk_size : `int`, Optional (Keyword only)
        The maximal size of a raw chunk to output. Defaults to `1000`.
    sanitize : `bool`, Optional (Keyword only)
        Whether the output stream should be sanitized. Defaults to `False`.
        
        > Sanitization is only applied after an output chunk is cut to size, so it is recommended to not set
        `chunk_size` over `1000` if `sanitize` is given as `True`.
    
    Returns
    -------
    output_stream : ``ChannelOutputStream``
    """
    return ChannelOutputStream(client, channel, chunk_size, sanitize)


def get_channel_stdin(client, channel, *, check=None):
    """
    Gets an input stream from the given channel.
    
    Parameters
    ----------
    client : ``Client``
        The client who redirects the stream.
    channel : ``ChannelText``
        The channel from where the stream will be redirected form.
    check : `None` or `callable`, Optional (Keyword only)
        Check which message's content should be feed to the input. Defaults to `None`.
        
        Should accept the following parameters:
        
        +-----------+---------------+
        | Name      | Type          |
        +===========+===============+
        | message   | ``Message``   |
        +-----------+---------------+
        
        Should return the following values:
        
        +-------------------+---------------+
        | Name              | Type          |
        +===================+===============+
        | should_redirect   | ``bool``      |
        +-------------------+---------------+
    
    Returns
    -------
    input_stream : ``ChannelInputStream``
    """
    return ChannelInputStream(client, channel, check)



__all__ = ('MessageInteraction',)

from ...backend.export import include

from ..bases import DiscordEntity
from ..user import User

InteractionType = include('InteractionType')


class MessageInteraction(DiscordEntity):
    """
    Sent with a ``Message``, when the it is a response to an ``InteractionEvent``.
    
    Attributes
    ----------
    id : `int`
        The interaction's identifier.
    name : `str`
        The invoked interaction's name.
    type : ``InteractionType``
        The interaction's type.
    user : ``ClientUserBase``
        Who invoked the interaction.
    """
    __slots__ = ('name', 'type', 'user')
    def __new__(cls, data):
        """
        Creates a new ``MessageInteraction`` instance from the received data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Message interaction data.
        """
        self = object.__new__(cls)
        self.id = int(data['id'])
        self.name = data['name']
        self.type = InteractionType.get(data['type'])
        self.user = User(data['user'])
        
        return self
    
    
    def __repr__(self):
        """Returns the message interaction's representation."""
        repr_parts = ['<', self.__class__.__name__, ' id=', repr(self.id), ', type=']
        
        interaction_type = self.type
        repr_parts.append(interaction_type.name)
        repr_parts.append(' (')
        repr_parts.append(repr(interaction_type.value))
        repr_parts.append(')')
        
        repr_parts.append(', name=')
        repr_parts.append(repr(self.name))
        repr_parts.append('>')
        
        return ''.join(repr_parts)



__all__ = ('SlasherCommandError', 'SlasherApplicationCommandParameterConversionError')

from random import choice

from ...backend.utils import copy_docs
from ...backend.export import include
from ...backend.analyzer import CallableAnalyzer

from ...discord.interaction import InteractionType
from ...discord.exceptions import DiscordException, ERROR_CODES

SlasherApplicationCommand = include('SlasherApplicationCommand')
Slasher = include('Slasher')

class SlasherCommandError(Exception):
    """
    Base class for slash command internal errors.
    """
    pass

    @property
    def pretty_repr(self):
        """
        Returns the pretty representation of the exception.
        
        Returns
        -------
        representation : `str`
        """
        return ''

class SlasherApplicationCommandParameterConversionError(SlasherCommandError):
    """
    Exception raised when a command's parameter's parsing fails.
    
    Attributes
    ----------
    _pretty_repr : `None` or `str`
        generated pretty representation of the exception.
    _repr : `None` or `str`
        The generated error message.
    parameter_name : `None` or `str`
        The parameter's name, which failed to be parsed.
    received_value : `None` or `str`
        The parameter's received value.
    excepted_type : `None` or `str`
        The parameter's expected type's name.
    expected_values : `None` or `list` of `Any`
        Expected values.
    """
    def __init__(self, parameter_name, received_value, excepted_type, expected_values):
        """
        Creates a new ``SlasherApplicationCommandParameterConversionError`` instance with the given parameters.
        
        Parameters
        ----------
        parameter_name : `None` or `str`
            The parameter's name, which failed to be parsed.
        received_value : `None` or `str`
            The parameter's received value.
        excepted_type : `None` or `str`
            The parameter's expected type's name.
        expected_values : `None` or `list` of `Any`
            Expected values.
        """
        self.parameter_name = parameter_name
        self.received_value = received_value
        self.excepted_type = excepted_type
        self.expected_values = expected_values
        self._repr = None
        self._pretty_repr = None
        Exception.__init__(self, parameter_name, received_value, excepted_type, expected_values)
    
    def __repr__(self):
        """Returns the representation of the parameter conversion error."""
        repr_ = self._repr
        if repr_ is None:
            repr_ = self._create_repr()
        
        return repr_

    def _create_repr(self):
        """
        Creates the representation of the parsing syntax error.
        
        Returns
        -------
        repr_ : `str`
            The representation of the syntax error.
        """
        repr_parts = [self.__class__.__name__]
        
        parameter_name = self.parameter_name
        if (parameter_name is not None):
            repr_parts.append('\n')
            repr_parts.append('parameter name: ')
            repr_parts.append(repr(parameter_name))
        
        excepted_type = self.excepted_type
        if (excepted_type is not None):
            repr_parts.append(
                '\n'
                'expected type: '
            )
            repr_parts.append(excepted_type)
        
        expected_values = self.expected_values
        if (expected_values is not None):
            repr_parts.append(
                '\n'
                'expected value(s):'
            )
            
            index = 0
            limit = len(expected_values)
            while True:
                value = expected_values[index]
                index += 1
                
                repr_parts.append(repr(value))
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
        
        repr_parts.append(
            '\n'
            'received value: '
        )
        received_value = self.received_value
        if (received_value is None):
            repr_parts.append('N/A')
        else:
            repr_parts.append(repr(received_value))
        
        repr_ = ''.join(repr_parts)
        self._repr = repr_
        return repr_
    
    
    @property
    @copy_docs(SlasherCommandError.pretty_repr)
    def pretty_repr(self):
        pretty_repr = self._pretty_repr
        if pretty_repr is None:
            pretty_repr = self._create_pretty_repr()
        
        return pretty_repr
    
    
    def _create_pretty_repr(self):
        """
        Creates the pretty representation of the parameter conversion error.
        
        Returns
        -------
        repr_ : `str`
            The representation of the parameter conversion error.
        """
        repr_parts = ['Parameter conversion failed\n']
        
        parameter_name = self.parameter_name
        if (parameter_name is not None):
            repr_parts.append('\n')
            repr_parts.append('Name: `')
            repr_parts.append(parameter_name)
            repr_parts.append('`')
        
        excepted_type = self.excepted_type
        if (excepted_type is not None):
            repr_parts.append(
                '\n'
                'Excepted type: `'
            )
            repr_parts.append(excepted_type)
            repr_parts.append('`')
    
        
        expected_values = self.expected_values
        if (expected_values is not None):
            repr_parts.append(
                '\n'
                'Expected value(s):'
            )
            
            for expected_value in expected_values:
                repr_parts.append('\n- `')
                repr_parts.append(expected_value)
                repr_parts.append('`')
        
        repr_parts.append(
            '\n'
            'Received: '
        )
        received_value = self.received_value
        if (received_value is None):
            repr_parts.append('N/A')
        else:
            repr_parts.append('`')
            repr_parts.append(received_value)
            repr_parts.append('`')
        
        pretty_repr = ''.join(repr_parts)
        self._pretty_repr = pretty_repr
        return pretty_repr



ERROR_MESSAGES = [
    (
       'Exception occurred meanwhile processing your interaction.\n'
       'Our highly educated Cirno-s are already working on the problem.'
    ), (
        'Your command\'s execution failed.\n'
        'We are putting Reimu to investigate about the issue.'
    ), (
        'The command died out.\n'
        'Maybe Izaoyi love-shop has some antidote.'
    ), (
        'Something went wrong.\n'
        'Do you need driver license for broom riding?'
    ),
]

def default_slasher_random_error_message_getter():
    """
    Returns random error message getter of ``Slasher``.
    
    Returns
    -------
    error_message : `str`
    """
    return choice(ERROR_MESSAGES)


def _validate_random_error_message_getter(random_error_message_getter):
    """
    Validates the given `random_error_message_getter`. It should accept no parameters and return a string.
    
    Parameters
    ----------
    random_error_message_getter : `callable`
        The random message getter to validate.
    
    Raises
    ------
    TypeError
        - If `random_error_message_getter` is not callable.
        - If `random_error_message_getter` is not async.
        - If `random_error_message_getter` excepts not `0` parameters.
    """
    analyzer = CallableAnalyzer(random_error_message_getter)
    if analyzer.is_async():
        raise TypeError(f'`random_error_message_getter` cannot be async, got {random_error_message_getter!r}.')
    
    min_, max_ = analyzer.get_non_reserved_positional_parameter_range()
    if (min_ > 0):
        raise TypeError(f'A `{random_error_message_getter}` should accept `0` parameters, meanwhile '
            f'{random_error_message_getter!r} accepts between `{min_}` and `{max_}`.')


async def default_slasher_exception_handler(client, interaction_event, command, exception):
    """
    Default ``Slasher`` exception handler.
    
    This function is a coroutine.
    
    Parameters
    ----------
    client : ``Client``
        The respective client.
    interaction_event : ``InteractionEvent``
        The received interaction event.
    command : ``SlasherApplicationCommand`` or ``ComponentCommand``
        The command, which raised.
    exception : `BaseException`
        The occurred exception.
    
    Returns
    -------
    handled : `bool`
        Whether the error handler handled the exception.
    """
    if isinstance(exception, SlasherCommandError):
        forward = exception.pretty_repr
        render = False
    elif isinstance(exception, DiscordException) and (exception.status == 500):
        forward = None
        render = True
    elif (interaction_event.type is InteractionType.application_command) and interaction_event.is_unanswered():
        forward = client.slasher._random_error_message_getter()
        render = True
    else:
        forward = None
        render = True
    
    if (forward is not None):
        try:
            await client.interaction_response_message_create(interaction_event, forward,
                show_for_invoking_user_only=True)
        except BaseException as err:
            if isinstance(err, ConnectionError):
                pass
            elif isinstance(err, DiscordException) and (
                    (err.status == 500) or (err.code == ERROR_CODES.unknown_interaction)
            ):
                pass
            else:
                raise
    
    if render:
        await _render_application_command_exception(client, command, exception)
    
    return True


async def _render_application_command_exception(client, command, exception):
    """
    Renders interaction command exception.
    
    Parameters
    ----------
    client : ``Client``
        The respective client.
    command : ``SlasherApplicationCommand`` or ``ComponentCommand``
        The command, which raised.
    exception : `BaseException`
        The occurred exception.
    """
    if isinstance(command, SlasherApplicationCommand):
        command_name = command.name
    else:
        command_name = command.__class__.__name__
    
    await client.events.error(client, f'`Slasher` while calling `{command_name}`', exception)


async def handle_command_exception(entity, client, interaction_event, exception):
    """
    handles slash command exception.
    
    This function is a coroutine.

    Parameters
    ----------
    entity : ``ComponentCommand``, ``Slasher``, ``SlasherApplicationCommand``, ``SlasherApplicationCommandCategory``,
            ``SlasherApplicationCommandFunction``, ``SlasherApplicationCommandParameterAutoCompleter``
        The entity to iterate it's exception handlers.
    client : ``Client``
        The respective client.
    interaction_event : ``InteractionEvent``
        The received interaction event.
    exception : `BaseException`
        The occurred exception.
    """
    for exception_handler in _iter_exception_handlers(entity):
        try:
            handled = await exception_handler(client, interaction_event, entity, exception)
        except BaseException as err:
            await client.events.error(client, 'handle_command_exception', err)
            return
        
        if handled:
            return
    
    if not isinstance(exception, SlasherCommandError):
        await _render_application_command_exception(client, entity, exception)


def test_exception_handler(exception_handler):
    """
    Tests whether the given exception handler accepts the expected amount of parameters.
    
    Parameters
    ----------
    exception_handler : `CoroutineFunction`
        A function, which handles an exception and returns whether handled it.
        
        The following parameters are passed to it:
        
        +-------------------+-------------------------------------------------------------------------------+
        | Name              | Type                                                                          |
        +===================+===============================================================================+
        | client            | ``Client``                                                                    |
        +-------------------+-------------------------------------------------------------------------------+
        | interaction_event | ``InteractionEvent``                                                          |
        +-------------------+-------------------------------------------------------------------------------+
        | command           | ``ComponentCommand``, ``SlasherApplicationCommand``,                          |
        |                   | ``SlasherApplicationCommandFunction``, ``SlasherApplicationCommandCategory``, |
        |                   | ``SlasherApplicationCommandParameterAutoCompleter``                           |
        +-------------------+-------------------------------------------------------------------------------+
        | exception         | `BaseException`                                                               |
        +-------------------+-------------------------------------------------------------------------------+
        
        Should return the following parameters:
        
        +-------------------+-----------+
        | Name              | Type      |
        +===================+===========+
        | handled           | `bool`    |
        +-------------------+-----------+
    
    Raises
    ------
    TypeError
        - If `exception_handler` accepts bad amount of parameters.
        - If `exception_handler` is not a coroutine function.
    """
    analyzer = CallableAnalyzer(exception_handler)
    if not analyzer.is_async():
        raise TypeError('`exception_handler` should be given as `async` function.')
    
    min_, max_ = analyzer.get_non_reserved_positional_parameter_range()
    if min_ > 4:
        raise TypeError(f'`exception_handler` should accept `4` parameters, meanwhile the given callable expects at '
            f'least `{min_!r}`, got `{exception_handler!r}`.')
    
    if min_ != 4:
        if max_ < 4:
            if not analyzer.accepts_args():
                raise TypeError(f'`exception_handler` should accept `4` parameters, meanwhile the given callable '
                    f'expects up to `{max_!r}`, got `{exception_handler!r}`.')


def _register_exception_handler(parent, first, exception_handler):
    """
    Registers an exception handler to the respective entity. This function is wrapped inside of
    `functools.partial` with it's `parent` and `first` parameters.
    
    Parameters
    ----------
    parent : ``Slasher``, ``SlasherApplicationCommand`, ``SlasherApplicationCommandCategory``, ``ComponentCommand``,
            ``SlasherApplicationCommandFunction``, ``SlasherApplicationCommandParameterAutoCompleter``
        The slasher to register the exception handler to.
    first : `bool`
        Whether the exception handler should run first.
    exception_handler : `CoroutineFunction`
        Exception handler to register.
    
    Returns
    -------
    exception_handler : `CoroutineFunction`
    
    Raises
    ------
    RuntimeError
        If `exception_handler` is given as `None`.
    """
    if (exception_handler is None):
        raise RuntimeError('`exception_handler` cannot be `None`.')
    
    return parent._register_exception_handler(exception_handler, first)


def _iter_exception_handlers(entity):
    """
    Iterates the given entity's exception handlers.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    entity : ``ComponentCommand``, ``Slasher``, ``SlasherApplicationCommand``, ``SlasherApplicationCommandCategory``,
            ``SlasherApplicationCommandFunction``, ``SlasherApplicationCommandParameterAutoCompleter``
        The entity to iterate it's exception handlers.
    
    Yields
    ------
    exception_handler : `CoroutineFunction`
    """
    while True:
        exception_handlers = entity._exception_handlers
        if (exception_handlers is not None):
            yield from exception_handlers
        
        if isinstance(entity, Slasher):
            break
        
        parent_reference = entity._parent_reference
        if parent_reference is None:
            break
        
        entity = parent_reference()
        if entity is None:
            break



__all__ = ('DefaultAvatar', 'FriendRequestFlag', 'HypesquadHouse', 'PremiumType', 'RelationshipType', 'Status',
    'Theme', )

from ...backend.export import export

from ..bases import PreinstancedBase, Preinstance as P
from ..color import Color

from ..http import urls as module_urls


class DefaultAvatar(PreinstancedBase):
    """
    Represents a default avatar of a user. Default avatar is used, when the user has no avatar set.
    
    There are some predefined default avatars and there should be no more instances created.
    
    +-----------+-----------+-----------+
    | name      | value     | color     |
    +===========+===========+===========+
    | blue      | 0         | 0x7289da  |
    +-----------+-----------+-----------+
    | gray      | 1         | 0x747f8d  |
    +-----------+-----------+-----------+
    | green     | 2         | 0x43b581  |
    +-----------+-----------+-----------+
    | orange    | 3         | 0xfaa61a  |
    +-----------+-----------+-----------+
    | red       | 4         | 0xf04747  |
    +-----------+-----------+-----------+
    
    Attributes
    ----------
    color : ``Color``
        The color of the default avatar.
    name : `str`
        The name of the default avatar's color.
    value : ``int`
        The identifier value of the default avatar.
        
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``DefaultAvatar``) objects
        The predefined default avatar instances stored for lookup.
    VALUE_TYPE : `type` = `int`
        The default avatars' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the default avatars.
    DEFAULT_COLOR : ``Color`` = `Color()`
        The default color of the default avatars.
    """
    INSTANCES = {}
    VALUE_TYPE = int
    DEFAULT_COLOR = Color()
    
    __slots__ = ('color',)
    
    @classmethod
    def for_(cls, user):
        """
        Returns the default avatar for the given user.
        
        Parameters
        ----------
        user : ``UserBase`` instance
            The user, who's default avatar will be looked up.

        Returns
        -------
        default_avatar : ``DefaultAvatar``
        """
        INSTANCES = cls.INSTANCES
        return INSTANCES[user.discriminator%len(INSTANCES)]
    
    @classmethod
    def _from_value(cls, value):
        """
        Creates a new default avatar with the given value.
        
        Parameters
        ----------
        value : `int`
            The identifier value of the default avatar.
        
        Returns
        -------
        self : ``DefaultAvatar``
            The created default avatar.
        """
        self = object.__new__(cls)
        self.value = value
        self.name = cls.DEFAULT_NAME
        self.color = self.DEFAULT_COLOR
        return self
    
    def __init__(self, value, name, color):
        """
        Creates a default avatar and puts it into the class's `.INSTANCES`.
        
        Parameters
        ----------
        color : ``Color``
            The color of the default avatar.
        name : `str`
            The name of the default avatar's color.
        value : ``int`
            The identifier value of the default avatar.
        """
        self.value = value
        self.name = name
        self.color = color
        self.INSTANCES[value] = self
    
    def __repr__(self):
        """Returns the default's avatar's representation."""
        return f'<{self.__class__.__name__} name={self.name}, value={self.value}, color={self.color!r}>'
    
    url = property(module_urls.default_avatar_url)
    
    
    # predefined
    blue = P(0, 'blue', Color(0x7289da))
    gray = P(1, 'gray', Color(0x747f8d))
    green = P(2, 'green', Color(0x43b581))
    orange = P(3, 'orange', Color(0xfaa61a))
    red = P(4, 'red', Color(0xf04747))



class HypesquadHouse(PreinstancedBase):
    """
    Represents Discord's hypesquad house.
    
    Attributes
    ----------
    value : `int`
        The Discord side identifier value of the hypesquad house.
    name : `str`
        The default name of the hypesquad house.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``HypesquadHouse``) items
        Stores the predefined hypesquad houses. This container is accessed when converting Discord side hypesquad
        house's value to it's wrapper side representation.
    VALUE_TYPE : `type` = `int`
        The hypesquad houses' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the hypesquad houses.
    
    Every predefined hypesquad house can also be accessed as class attribute:
    
    +-----------------------+---------------+-------+
    | Class attribute name  | name          | value |
    +=======================+===============+=======+
    | none                  | none          | 0     |
    +-----------------------+---------------+-------+
    | bravery               | bravery       | 1     |
    +-----------------------+---------------+-------+
    | brilliance            | brilliance    | 2     |
    +-----------------------+---------------+-------+
    | balance               | balance       | 3     |
    +-----------------------+---------------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    
    __slots__ = ()

    # predefined
    none = P(0, 'none')
    bravery = P(1, 'bravery')
    brilliance = P(2, 'brilliance')
    balance = P(3, 'balance')


class Status(PreinstancedBase):
    """
    Represents a Discord user's status.
    
    Attributes
    ----------
    position : `int`
        Internal position of the status for sorting purposes.
    value : `str`
        The identifier value of the status.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`str`, ``Status``) items
        A container what stores the predefined statuses in `value` - `status` relation. This container is accessed
        when translating status value to ``Status`` object.
    VALUE_TYPE : `type` = `str`
        The statuses' values' type.
    DEFAULT_NAME : `str` = `''`
        The default name of the statuses. Statuses sue their value as names, so this field is not used.
    DEFAULT_POSITION : `int` = `5`
        The default position of the statuses'.
    
    Each predefined status also can be accessed as a class attribute:
    
    +-----------------------+-----------+-----------+
    | Class attribute name  | position  | value     |
    +=======================+===========+===========+
    | online                | 0         | idle      |
    +-----------------------+-----------+-----------+
    | idle                  | 1         | idle      |
    +-----------------------+-----------+-----------+
    | dnd                   | 2         | dnd       |
    +-----------------------+-----------+-----------+
    | offline               | 3         | offline   |
    +-----------------------+-----------+-----------+
    | invisible             | 3         | invisible |
    +-----------------------+-----------+-----------+
    """
    INSTANCES = {}
    VALUE_TYPE = str
    DEFAULT_NAME = ''
    DEFAULT_POSITION = 4
    
    __slots__ = ('position', )
    
    def __init__(self, value, name, position):
        """
        Creates a new status and stores it at the class's `.INSTANCES` class attribute.
        
        Parameters
        ----------
        value : `str`
            The identifier value of the status.
        position : `int`
            Internal position of the status for sorting purposes.
        """
        self.value = value
        self.name = name
        self.position = position
        self.INSTANCES[value] = self
    
    @classmethod
    def _from_value(cls, value,):
        """
        Creates a new status object from the given value.
        
        Parameters
        ----------
        value : ``.VALUE_TYPE``
            The value what has no representation yet.
        
        Returns
        -------
        self : ``Status`` instance
            The created status.
        """
        self = object.__new__(cls)
        self.value = value
        self.name = value
        self.position = cls.DEFAULT_POSITION
        self.INSTANCES[value] = self
        return self
    
    def __repr__(self):
        """Returns the representation of the status."""
        return f'<{self.__class__.__name__} value={self.value!r}>'
    
    def __gt__(self, other):
        """Returns whether this status's position is greater than the other's."""
        other_type = other.__class__
        self_type = self.__class__
        if other_type is other_type:
            pass
        elif issubclass(other_type, self_type.VALUE_TYPE):
            try:
                other = self_type.INSTANCES[other]
            except KeyError:
                return P
        else:
            return P
        
        if self.position > other.position:
            return True
        else:
            return False
    
    def __ge__(self,other):
        """Returns whether this status's position is greater than the other's or whether the two status is equal."""
        other_type = other.__class__
        self_type = self.__class__
        if other_type is other_type:
            pass
        elif issubclass(other_type, self_type.VALUE_TYPE):
            try:
                other = self_type.INSTANCES[other]
            except KeyError:
                return P
        else:
            return P
        
        self_position = self.position
        other_position = other.position
        if self_position > other_position:
            return True
        
        if self_position < other_position:
            return False
        
        if self.value == other.value:
            return True
        
        return False
    
    def __eq__(self, other):
        """Returns whether the two status is equal."""
        other_type = other.__class__
        self_type = self.__class__
        if other_type is other_type:
            pass
        elif issubclass(other_type, self_type.VALUE_TYPE):
            try:
                other = self_type.INSTANCES[other]
            except KeyError:
                return P
        else:
            return P
    
        if self.position != other.position:
            return False
        
        if self.value == other.value:
            return True
        
        return False
    
    def __ne__(self,other):
        """Returns whether the two status is not equal."""
        other_type = other.__class__
        self_type = self.__class__
        if other_type is other_type:
            pass
        elif issubclass(other_type, self_type.VALUE_TYPE):
            try:
                other = self_type.INSTANCES[other]
            except KeyError:
                return P
        else:
            return P
        
        if self.position != other.position:
            return True
            
        if self.value == other.value:
            return False
        
        return True
    
    def __le__(self, other):
        """Returns whether this status's position is less than the other's or whether the two status is equal."""
        other_type = other.__class__
        self_type = self.__class__
        if other_type is other_type:
            pass
        elif issubclass(other_type, self_type.VALUE_TYPE):
            try:
                other = self_type.INSTANCES[other]
            except KeyError:
                return P
        else:
            return P
        
        self_position = self.position
        other_position = other.position
        if self_position < other_position:
            return True
        
        if self_position > other_position:
            return False
        
        if self.value == other.value:
            return True
        
        return False
    
    def __lt__(self,other):
        """Returns whether this status's position is less than the other's."""
        other_type = other.__class__
        self_type = self.__class__
        if other_type is other_type:
            pass
        elif issubclass(other_type, self_type.VALUE_TYPE):
            try:
                other = self_type.INSTANCES[other]
            except KeyError:
                return P
        else:
            return P
        
        if self.position < other.position:
            return True
        
        return False
    
    # predefined
    online = P('online', 'online', 0)
    idle = P('idle', 'idle', 1)
    dnd = P('dnd', 'dnd', 2)
    offline = P('offline', 'offline', 3)
    invisible = P('invisible', 'invisible', 3)


class PremiumType(PreinstancedBase):
    """
    Represents Discord's premium types.
    
    Attributes
    ----------
    name : `str`
        The default name of the premium type.
    value : `int`
        The Discord side identifier value of the premium type.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``PremiumType``) items
        A container what stores the predefined premium types and also is accessed when translating their identifier
        value to their representation.
    VALUE_TYPE : `type` = `int`
        The premium types' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the premium types.
    
    Each predefined premium type can also be accessed as class attribute:
    
    +-----------------------+---------------+-------+
    | Class attribute name  | name          | value |
    +=======================+===============+=======+
    | none                  | none          | 0     |
    +-----------------------+---------------+-------+
    | nitro_classic         | nitro_classic | 1     |
    +-----------------------+---------------+-------+
    | nitro                 | nitro         | 2     |
    +-----------------------+---------------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    
    __slots__ = ()
    
    # predefined
    none = P(0, 'none')
    nitro_classic = P(1, 'nitro_classic')
    nitro = P(2, 'nitro')


@export
class RelationshipType(PreinstancedBase):
    """
    Represents a ``Relationship``'s type.
    
    Attributes
    ----------
    name : `str`
        The relationship type's name.
    value : `int`
        The Discord side identifier value of the relationship type.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``RelationshipType``) items
        The predefined relation types stored in a list, so they can be accessed with their respective value as key.
        This behaviour is used to translate their Discord side value to their representation.
    VALUE_TYPE : `type` = `int`
        The relationship types' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the relation types.
    
    Each predefined relationship type can also be accessed as class attribute:
    
    +-----------------------+-------------------+-------+
    | Class attribute name  | name              | value |
    +=======================+===================+=======+
    | stranger              | stranger          | 0     |
    +-----------------------+-------------------+-------+
    | friend                | friend            | 1     |
    +-----------------------+-------------------+-------+
    | blocked               | blocked           | 2     |
    +-----------------------+-------------------+-------+
    | pending_incoming      | pending_incoming  | 3     |
    +-----------------------+-------------------+-------+
    | pending_outgoing      | pending_outgoing  | 4     |
    +-----------------------+-------------------+-------+
    | implicit              | implicit          | 5     |
    +-----------------------+-------------------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    
    __slots__ = ()
    
    # predefined
    stranger = P(0, 'stranger')
    friend = P(1, 'friend')
    blocked = P(2, 'blocked')
    pending_incoming = P(3, 'pending_incoming')
    pending_outgoing = P(4, 'pending_outgoing')
    implicit = P(5, 'implicit')


class FriendRequestFlag(PreinstancedBase):
    """
    Represents the friend request flags of a user.
    
    Attributes
    ----------
    name : `str`
        The default name of the friend request flag.
    value : `int`
        Internal identifier value of the friend request flag used for lookup.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``FriendRequestFlag``) items
        A container to store the predefined friend request flags. This container is accessed by ``.get`` what
        translates the friend request flags' value to their wrapper side representation.
    VALUE_TYPE : `type` = `int`
        The friend request flags' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the friend request flags.
    
    Every predefined friend request flag can also be accessed as class attribute:
    
    +---------------------------+---------------------------+-------+
    | Class attribute name      | name                      | value |
    +===========================+===========================+=======+
    | none                      | none                      | 0     |
    +---------------------------+---------------------------+-------+
    | mutual_guilds             | mutual_guilds             | 1     |
    +---------------------------+---------------------------+-------+
    | mutual_friends            | mutual_friends            | 2     |
    +---------------------------+---------------------------+-------+
    | mutual_guilds_and_friends | mutual_guilds_and_friends | 3     |
    +---------------------------+---------------------------+-------+
    | all                       | all                       | 4     |
    +---------------------------+---------------------------+-------+
    """
    # class related
    INSTANCES = {}
    VALUE_TYPE = int
    
    __slots__ = ()
    
    @classmethod
    def get(cls, data):
        """
        Converts the friend request flag data sent by Discord to it's wrapper side representation.
        
        Parameters
        ----------
        data : `dict` of (`str`, `bool`)
            Received friend request flag data.
        
        Returns
        -------
        friend_request_flag : ``FriendRequestFlag``
        """
        if data is None:
            return cls.none
        
        all_ = data.get('all', False)
        if all_:
            key = 4
        else:
            mutual_guilds = data.get('mutual_guilds', False)
            mutual_friends = data.get('mutual_friends', False)
            
            key = mutual_guilds+(mutual_friends<<1)
        
        return cls.INSTANCES[key]
    
    
    def encode(self):
        """
        Returns the friend request flag's Discord side representation.
        
        Returns
        -------
        result : `dict` of (`str`, `bool`) items
        """
        value = self.value
        result = {}
        if (value>>2)&1:
            result['all'] = True
        else:
            if (value>>1)&1:
                result['mutual_friends'] = True
            if value&1:
                result['mutual_guilds'] = True
        
        return result
    
    # predefined
    none = P(0, 'none')
    mutual_guilds = P(1, 'mutual_guilds')
    mutual_friends = P(2, 'mutual_friends')
    mutual_guilds_and_friends = P(3, 'mutual_guilds_and_friends')
    all = P(4, 'all')


class Theme(PreinstancedBase):
    """
    Represents a user's theme.
    
    Attributes
    ----------
    value : `str`
        The discord side identifier value of the theme.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`str`, ``Theme``) items
        Stores the predefined themes in `value` - `theme` relation. This container is accessed when converting a theme
        to it's representation.
    VALUE_TYPE : `type` = `str`
        The themes' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the themes.
    
    Each predefined theme instance can also be accessed as class attribute:
    
    +-----------------------+-----------+
    | Class attribute name  | value     |
    +=======================+===========+
    | dark                  | dark      |
    +-----------------------+-----------+
    | light                 | light     |
    +-----------------------+-----------+
    """
    INSTANCES = {}
    VALUE_TYPE = str
    
    __slots__ = ()
    
    # predefined
    dark = P('dark', 'dark')
    light = P('light', 'light')
